# DatabaseUpgrade.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2006, 2007, 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details
"""Handles the heavy lifting of upgrading the database.

   Holds methods to copy database contents around, utility classes
   to talk to old database versions, and so forth.
   """

# pylint: disable-msg=E0611
# sqlobject confuses pylint here
from sqlobject import sqlhub, SQLObject, IntCol, UnicodeCol, RelatedJoin, \
        EnumCol, MultipleJoin, connectionForURI, ForeignKey, BoolCol, \
        SQLObjectNotFound
# pylint: enable-msg=E0611
from logging import Logger
from sutekh.core.SutekhObjects import PhysicalCard, AbstractCard, \
        PhysicalCardSet, Expansion, Clan, Virtue, Discipline, Rarity, \
        RarityPair, CardType, Ruling, aObjectList, DisciplinePair, Creed, \
        Sect, Title, flush_cache
from sutekh.core.CardSetHolder import CachedCardSetHolder
from sutekh.SutekhUtility import refresh_tables
from sutekh.core.DatabaseVersion import DatabaseVersion
from sutekh.core.Abbreviations import Rarities

# This file handles all the grunt work of the database upgrades. We have some
# (arguablely overly) complex trickery to read old databases, and we create a
# copy in sqlite memory database first, before commiting to the actual DB

# Utility Exception
class UnknownVersion(Exception):
    """Exception for versions we cannot handle"""
    def __init__(self, sTableName):
        Exception.__init__(self)
        self.sTableName = sTableName

    def __str__(self):
        return "Unrecognised version for %s" % self.sTableName

# We Need to clone the SQLObject classes in SutekhObjects so we can read
# old versions
# pylint: disable-msg=C0103, W0232
# C0103 - names set largely by SQLObject conventions, so ours don't apply
# W0232 - SQLObject classes don't have user defined __init__
class PhysicalCardSet_v3(SQLObject):
    """PhysicalCardSet V3 - doesn't have the inuse flag."""
    class sqlmeta:
        """meta class used to set the correct table."""
        table = PhysicalCardSet.sqlmeta.table
    name = UnicodeCol(alternateID=True, length=50)
    author = UnicodeCol(length=50, default='')
    comment = UnicodeCol(default='')
    annotations = UnicodeCol(default='')
    cards = RelatedJoin('PhysicalCard', intermediateTable='physical_map',
            createRelatedTable=False)

class AbstractCard_v2(SQLObject):
    """AbstractCard V2 - doesn't have the burnoption"""
    class sqlmeta:
        """meta class used to set the correct table."""
        table = AbstractCard.sqlmeta.table
    canonicalName = UnicodeCol(alternateID=True, length=50)
    name = UnicodeCol(length=50)
    text = UnicodeCol()
    group = IntCol(default=None, dbName='grp')
    capacity = IntCol(default=None)
    cost = IntCol(default=None)
    life = IntCol(default=None)
    costtype = EnumCol(enumValues=['pool', 'blood', 'conviction', None],
            default=None)
    level = EnumCol(enumValues=['advanced', None], default=None)
    discipline = RelatedJoin('DisciplinePair',
            intermediateTable='abs_discipline_pair_map',
            createRelatedTable=False)
    rarity = RelatedJoin('RarityPair', intermediateTable='abs_rarity_pair_map',
            createRelatedTable=False)
    clan = RelatedJoin('Clan', intermediateTable='abs_clan_map',
            createRelatedTable=False)
    cardtype = RelatedJoin('CardType', intermediateTable='abs_type_map',
            createRelatedTable=False)
    sect = RelatedJoin('Sect', intermediateTable='abs_sect_map',
            createRelatedTable=False)
    title = RelatedJoin('Title', intermediateTable='abs_title_map',
            createRelatedTable=False)
    creed = RelatedJoin('Creed', intermediateTable='abs_creed_map',
            createRelatedTable=False)
    virtue = RelatedJoin('Virtue', intermediateTable='abs_virtue_map',
            createRelatedTable=False)
    rulings = RelatedJoin('Ruling', intermediateTable='abs_ruling_map',
            createRelatedTable=False)
    sets = RelatedJoin('AbstractCardSet_v3', intermediateTable='abstract_map',
            createRelatedTable=False)
    physicalCards = MultipleJoin('PhysicalCard')

class Rarity_v1(SQLObject):
    """Rarity v1 - doesn't have shortname"""
    class sqlmeta:
        """meta class used to set the correct table."""
        table = Rarity.sqlmeta.table
    name = UnicodeCol(alternateID=True, length=20)

class RarityPair_Rv1(SQLObject):
    """RarityPior table that joins with Rarity_v1"""
    class sqlmeta:
        """meta class used to set the correct table."""
        table = RarityPair.sqlmeta.table
    expansion = ForeignKey('Expansion')
    rarity = ForeignKey('Rarity_v1')
    cards = RelatedJoin('AbstractCard_v2',
            intermediateTable='abs_rarity_pair_map', createRelatedTable=False)

class AbstractCardSet_ACv2(SQLObject):
    """AbstractCardSet table that joins with AbstractCard_v2"""
    class sqlmeta:
        """meta class used to set the correct table."""
        table = 'abstract_card_set'
    name = UnicodeCol(alternateID=True, length=50)
    author = UnicodeCol(length=50, default='')
    comment = UnicodeCol(default='')
    annotations = UnicodeCol(default='')
    # Provides a join to AbstractCard_v2, needed to read old DB
    cards = RelatedJoin('AbstractCard_v2',
            intermediateTable='abstract_map', createRelatedTable=False)

class AbstractCardSet_v3(SQLObject):
    """Old abstract card set table."""
    class sqlmeta:
        """meta class used to set the correct table."""
        table = 'abstract_card_set'
    name = UnicodeCol(alternateID=True, length=50)
    author = UnicodeCol(length=50, default='')
    comment = UnicodeCol(default='')
    annotations = UnicodeCol(default='')
    cards = RelatedJoin('AbstractCard_v3', intermediateTable='abstract_map',
            createRelatedTable=False)

class AbstractCard_v3(SQLObject):
    """Old abstract card with join to abstract card set."""
    class sqlmeta:
        """meta class used to set the correct table."""
        table = AbstractCard.sqlmeta.table
    canonicalName = UnicodeCol(alternateID=True, length=50)
    name = UnicodeCol(length=50)
    text = UnicodeCol()
    group = IntCol(default=None, dbName='grp')
    capacity = IntCol(default=None)
    cost = IntCol(default=None)
    life = IntCol(default=None)
    costtype = EnumCol(enumValues=['pool', 'blood', 'conviction', None],
            default=None)
    level = EnumCol(enumValues=['advanced', None], default=None)
    burnoption = BoolCol(default=False)
    discipline = RelatedJoin('DisciplinePair',
            intermediateTable='abs_discipline_pair_map',
            createRelatedTable=False)
    rarity = RelatedJoin('RarityPair',
            intermediateTable='abs_rarity_pair_map', createRelatedTable=False)
    clan = RelatedJoin('Clan',
            intermediateTable='abs_clan_map', createRelatedTable=False)
    cardtype = RelatedJoin('CardType', intermediateTable='abs_type_map',
            createRelatedTable=False)
    sect = RelatedJoin('Sect', intermediateTable='abs_sect_map',
            createRelatedTable=False)
    title = RelatedJoin('Title', intermediateTable='abs_title_map',
            createRelatedTable=False)
    creed = RelatedJoin('Creed', intermediateTable='abs_creed_map',
            createRelatedTable=False)
    virtue = RelatedJoin('Virtue', intermediateTable='abs_virtue_map',
            createRelatedTable=False)
    rulings = RelatedJoin('Ruling', intermediateTable='abs_ruling_map',
            createRelatedTable=False)
    sets = RelatedJoin('AbstractCardSet_v3', intermediateTable='abstract_map',
            createRelatedTable=False)
    physicalCards = MultipleJoin('PhysicalCard')

class PhysicalCardSet_v4(SQLObject):
    """Old PCS table without parent links."""
    class sqlmeta:
        """meta class used to set the correct table."""
        table = PhysicalCardSet.sqlmeta.table
    name = UnicodeCol(alternateID=True, length=50)
    author = UnicodeCol(length=50, default='')
    comment = UnicodeCol(default='')
    annotations = UnicodeCol(default='')
    inuse = BoolCol(default=False)
    cards = RelatedJoin('PhysicalCard', intermediateTable='physical_map',
            createRelatedTable=False)

class MapAbstractCardToAbstractCardSet_v3(SQLObject):
    """Old abstract card map to need so we can drop tables properly"""
    class sqlmeta:
        """meta class used to set the correct table."""
        table = 'abstract_map'
    abstractCard = ForeignKey('AbstractCard_v3', notNull=True)
    abstractCardSet = ForeignKey('AbstractCardSet_v3', notNull=True)

# pylint: enable-msg=C0103, W0232

def check_can_read_old_database(oConn):
    # pylint: disable-msg=R0912
    # This has to be a giant if .. elif statement to check all the classes
    # subdivision isn't really beneficial
    """Can we upgrade from this database version?

       Sutekh 0.5.x and 0.6.x can upgrade from the versions in Sutekh 0.4.x,
       but no earlier
       """
    oVer = DatabaseVersion()
    oVer.expire_cache()
    if not oVer.check_tables_and_versions([Rarity], [Rarity.tableversion],
            oConn) and not oVer.check_tables_and_versions([Rarity], [1],
                    oConn):
        raise UnknownVersion("Rarity")
    if not oVer.check_tables_and_versions([Expansion],
            [Expansion.tableversion], oConn):
        raise UnknownVersion("Expansion")
    if not oVer.check_tables_and_versions([Discipline],
            [Discipline.tableversion], oConn):
        raise UnknownVersion("Discipline")
    if not oVer.check_tables_and_versions([Clan], [Clan.tableversion], oConn):
        raise UnknownVersion("Clan")
    if not oVer.check_tables_and_versions([CardType], [CardType.tableversion],
            oConn):
        raise UnknownVersion("CardType")
    if not oVer.check_tables_and_versions([Creed], [Creed.tableversion],
            oConn):
        raise UnknownVersion("Creed")
    if not oVer.check_tables_and_versions([Virtue], [Virtue.tableversion],
            oConn):
        raise UnknownVersion("Virtue")
    if not oVer.check_tables_and_versions([Sect], [Sect.tableversion], oConn):
        raise UnknownVersion("Sect")
    if not oVer.check_tables_and_versions([Title], [Title.tableversion],
            oConn):
        raise UnknownVersion("Title")
    if not oVer.check_tables_and_versions([Ruling], [Ruling.tableversion],
            oConn):
        raise UnknownVersion("Ruling")
    if not oVer.check_tables_and_versions([DisciplinePair],
            [DisciplinePair.tableversion], oConn):
        raise UnknownVersion("DisciplinePair")
    if not oVer.check_tables_and_versions([RarityPair],
            [RarityPair.tableversion], oConn):
        raise UnknownVersion("RarityPair")
    if not oVer.check_tables_and_versions([AbstractCard],
            [AbstractCard.tableversion], oConn) \
            and not oVer.check_table_in_versions(AbstractCard, [2, 3]):
        raise UnknownVersion("AbstractCard")
    if not oVer.check_tables_and_versions([PhysicalCard],
            [PhysicalCard.tableversion], oConn):
        raise UnknownVersion("PhysicalCard")
    if not oVer.check_tables_and_versions([PhysicalCardSet],
            [PhysicalCardSet.tableversion], oConn) and not \
                    oVer.check_table_in_versions(PhysicalCardSet, [3, 4]):
        raise UnknownVersion("PhysicalCardSet")
    # NB: This test should go away for Sutekh 0.8, as we will no longer
    # support upgrading from DB version which have abstract card sets
    if not oVer.check_table_in_versions(AbstractCardSet_v3, [-1, 3]):
        raise UnknownVersion("AbstractCardSet")
    return True

def old_database_count(oConn):
    """Check number of items in old DB fro progress bars, etc."""
    oVer = DatabaseVersion()
    iCount = 12 # Card property tables
    if oVer.check_tables_and_versions([AbstractCard],
            [AbstractCard.tableversion], oConn):
        iCount += AbstractCard.select(connection=oConn).count()
    elif oVer.check_tables_and_versions([AbstractCard], [3], oConn):
        # We log once for each AC copied, and once for each set of PC's
        # created from the abstract cards
        iCount += 2 * AbstractCard_v3.select(connection=oConn).count()
    elif oVer.check_tables_and_versions([AbstractCard], [2], oConn):
        iCount += 2 * AbstractCard_v2.select(connection=oConn).count()
    if oVer.check_tables_and_versions([PhysicalCard],
            [PhysicalCard.tableversion], oConn):
        iCount += PhysicalCard.select(connection=oConn).count()
    if oVer.check_tables_and_versions([PhysicalCardSet],
            [PhysicalCardSet.tableversion], oConn):
        iCount += PhysicalCardSet.select(connection=oConn).count()
    elif oVer.check_tables_and_versions([PhysicalCardSet], [3], oConn):
        iCount += PhysicalCardSet_v3.select(connection=oConn).count()
    elif oVer.check_tables_and_versions([PhysicalCardSet], [4], oConn):
        iCount += PhysicalCardSet_v4.select(connection=oConn).count()
    if oVer.check_tables_and_versions([AbstractCardSet_v3], [3], oConn):
        iCount += AbstractCardSet_v3.select(connection=oConn).count()
    return iCount

# pylint: disable-msg=W0612
# W0612 - we have lots of unused variables from creating copies

def copy_rarity(oOrigConn, oTrans):
    """Copy rarity tables, assumings same version"""
    for oObj in Rarity.select(connection=oOrigConn):
        oCopy = Rarity(id=oObj.id, name=oObj.name,
                shortname=oObj.shortname, connection=oTrans)

def copy_old_rarity(oOrigConn, oTrans, oVer):
    """Copy rarity table, upgrading versions as needed"""
    if oVer.check_tables_and_versions([Rarity], [Rarity.tableversion],
            oOrigConn):
        copy_rarity(oOrigConn, oTrans)
    elif oVer.check_tables_and_versions([Rarity], [1], oOrigConn):
        for oObj in Rarity_v1.select(connection=oOrigConn):
            oCopy = Rarity(id=oObj.id, name=oObj.name,
                    shortname=Rarities.shortname(oObj.name),
                    connection=oTrans)
    else:
        return (False, ["Unknown Version for Rarity"])
    return (True, [])

def copy_expansion(oOrigConn, oTrans):
    """Copy expansion, assuming versions match"""
    for oObj in Expansion.select(connection=oOrigConn):
        oCopy = Expansion(id=oObj.id, name=oObj.name,
                shortname=oObj.shortname, connection=oTrans)

def copy_old_expansion(oOrigConn, oTrans, oVer):
    """Copy Expansion, updating as needed"""
    if oVer.check_tables_and_versions([Expansion], [Expansion.tableversion],
            oOrigConn):
        copy_expansion(oOrigConn, oTrans)
    else:
        return (False, ["Unknown Expansion Version"])
    return (True, [])

def copy_discipline(oOrigConn, oTrans):
    """Copy Discipline, assuming versions match"""
    for oObj in Discipline.select(connection=oOrigConn):
        oCopy = Discipline(id=oObj.id, name=oObj.name,
            fullname=oObj.fullname, connection=oTrans)

def copy_old_discipline(oOrigConn, oTrans, oVer):
    """Copy disciplines, upgrading as needed."""
    if oVer.check_tables_and_versions([Discipline], [Discipline.tableversion],
            oOrigConn):
        copy_discipline(oOrigConn, oTrans)
    else:
        return (False, ["Unknown Discipline version"])
    return (True, [])

def copy_clan(oOrigConn, oTrans):
    """Copy Clan, assuming database versions match"""
    for oObj in Clan.select(connection=oOrigConn):
        oCopy = Clan(id=oObj.id, name=oObj.name, shortname=oObj.shortname,
                connection=oTrans)

def copy_old_clan(oOrigConn, oTrans, oVer):
    """Copy clan, upgrading as needed."""
    if oVer.check_tables_and_versions([Clan], [Clan.tableversion], oOrigConn):
        copy_clan(oOrigConn, oTrans)
    else:
        return (False, ["Unknown Clan Version"])
    return (True, [])

def copy_creed(oOrigConn, oTrans):
    """Copy Creed, assuming versions match"""
    for oObj in Creed.select(connection=oOrigConn):
        oCopy = Creed(id=oObj.id, name=oObj.name, shortname=oObj.shortname,
                connection=oTrans)

def copy_old_creed(oOrigConn, oTrans, oVer):
    """Copy Creed, updating if needed"""
    if oVer.check_tables_and_versions([Creed], [Creed.tableversion],
            oOrigConn):
        copy_creed(oOrigConn, oTrans)
    else:
        return (False, ["Unknown Creed Version"])
    return (True, [])

def copy_virtue(oOrigConn, oTrans):
    """Copy Virtue, assuming versions match"""
    for oObj in Virtue.select(connection=oOrigConn):
        oCopy = Virtue(id=oObj.id, name=oObj.name, fullname=oObj.fullname,
                connection=oTrans)

def copy_old_virtue(oOrigConn, oTrans, oVer):
    """Copy Virtue, updating if needed"""
    if oVer.check_tables_and_versions([Virtue], [Virtue.tableversion],
            oOrigConn):
        copy_virtue(oOrigConn, oTrans)
    else:
        return (False, ["Unknown Virtue Version"])
    return (True, [])

def copy_card_type(oOrigConn, oTrans):
    """Copy CardType, assuming versions match"""
    for oObj in CardType.select(connection=oOrigConn):
        oCopy = CardType(id=oObj.id, name=oObj.name, connection=oTrans)

def copy_old_card_type(oOrigConn, oTrans, oVer):
    """Copy CardType, upgrading as needed"""
    if oVer.check_tables_and_versions([CardType], [CardType.tableversion],
            oOrigConn):
        copy_card_type(oOrigConn, oTrans)
    else:
        return (False, ["Unknown CardType Version"])
    return (True, [])

def copy_ruling(oOrigConn, oTrans):
    """Copy Ruling, assuming versions match"""
    for oObj in Ruling.select(connection=oOrigConn):
        oCopy = Ruling(id=oObj.id, text=oObj.text, code=oObj.code,
                url=oObj.url, connection=oTrans)

def copy_old_ruling(oOrigConn, oTrans, oVer):
    """Copy Ruling, upgrading as needed"""
    if oVer.check_tables_and_versions([CardType], [CardType.tableversion],
            oOrigConn):
        copy_ruling(oOrigConn, oTrans)
    else:
        return (False, ["Unknown Ruling Version"])
    return (True, [])

def copy_discipline_pair(oOrigConn, oTrans):
    """Copy DisciplinePair, assuming versions match"""
    for oObj in DisciplinePair.select(connection=oOrigConn):
        oCopy = DisciplinePair(id=oObj.id, level=oObj.level,
               discipline=oObj.discipline, connection=oTrans)

def copy_old_discipline_pair(oOrigConn, oTrans, oVer):
    """Copy DisciplinePair, upgrading if needed"""
    if oVer.check_tables_and_versions([DisciplinePair],
            [DisciplinePair.tableversion], oOrigConn):
        copy_discipline_pair(oOrigConn, oTrans)
    else:
        return (False, ["Unknown Discipline Version"])
    return (True, [])

def copy_rarity_pair(oOrigConn, oTrans):
    """Copy RairtyPair, assuming versions match"""
    for oObj in RarityPair.select(connection=oOrigConn):
        oCopy = RarityPair(id=oObj.id, expansion=oObj.expansion,
                rarity=oObj.rarity, connection=oTrans)

def copy_old_rarity_pair(oOrigConn, oTrans, oVer):
    """Copy RarityPair, upgrading as needed"""
    if oVer.check_tables_and_versions([RarityPair], [RarityPair.tableversion],
            oOrigConn) and oVer.check_tables_and_versions([Rarity],
                    [Rarity.tableversion], oOrigConn):
        copy_rarity_pair(oOrigConn, oTrans)
    elif oVer.check_tables_and_versions([RarityPair],
            [RarityPair.tableversion], oOrigConn) and \
            oVer.check_tables_and_versions([Rarity], [1], oOrigConn):
        for oObj in RarityPair_Rv1.select(connection=oOrigConn):
            oCopy = RarityPair(id=oObj.id, expansion=oObj.expansion,
                    rarity=oObj.rarity, connection=oTrans)
    else:
        return (False, ["Unknown RarityPair version"])
    return (True, [])

def copy_sect(oOrigConn, oTrans):
    """Copy Sect, assuming versions match"""
    for oObj in Sect.select(connection=oOrigConn):
        oCopy = Sect(id=oObj.id, name=oObj.name, connection=oTrans)

def copy_old_sect(oOrigConn, oTrans, oVer):
    """Copy Sect, updating if needed"""
    if oVer.check_tables_and_versions([Sect], [Sect.tableversion], oOrigConn):
        copy_sect(oOrigConn, oTrans)
    else:
        return (False, ["Unknown Sect Version"])
    return (True, [])

def copy_title(oOrigConn, oTrans):
    """Copy Title, assuming versions match"""
    for oObj in Title.select(connection=oOrigConn):
        oCopy = Title(id=oObj.id, name=oObj.name, connection=oTrans)

def copy_old_title(oOrigConn, oTrans, oVer):
    """Copy Title, updating if needed"""
    if oVer.check_tables_and_versions([Title], [Title.tableversion],
            oOrigConn):
        copy_title(oOrigConn, oTrans)
    else:
        return (False, ["Unknown Title Version"])
    return (True, [])

def copy_abstract_card(oOrigConn, oTrans, oLogger):
    """Copy AbstractCard, assuming versions match"""
    # pylint: disable-msg=E1101, R0912
    # E1101 - SQLObject confuses pylint
    # R0912 - need the branches for this
    for oCard in AbstractCard.select(connection=oOrigConn):
        oCardCopy = AbstractCard(id=oCard.id,
                canonicalName=oCard.canonicalName, name=oCard.name,
                text=oCard.text, connection=oTrans)
        # If I don't do things this way, I get encoding issues
        # I don't really feel like trying to understand why
        oCardCopy.group = oCard.group
        oCardCopy.capacity = oCard.capacity
        oCardCopy.cost = oCard.cost
        oCardCopy.costtype = oCard.costtype
        oCardCopy.level = oCard.level
        oCardCopy.life = oCard.life
        oCardCopy.burnoption = oCard.burnoption
        for oData in oCard.rarity:
            oCardCopy.addRarityPair(oData)
        for oData in oCard.discipline:
            oCardCopy.addDisciplinePair(oData)
        for oData in oCard.rulings:
            oCardCopy.addRuling(oData)
        for oData in oCard.clan:
            oCardCopy.addClan(oData)
        for oData in oCard.cardtype:
            oCardCopy.addCardType(oData)
        for oData in oCard.sect:
            oCardCopy.addSect(oData)
        for oData in oCard.title:
            oCardCopy.addTitle(oData)
        for oData in oCard.creed:
            oCardCopy.addCreed(oData)
        for oData in oCard.virtue:
            oCardCopy.addVirtue(oData)
        oCardCopy.syncUpdate()
        oLogger.info('copied AC %s', oCardCopy.name)

def copy_old_abstract_card(oOrigConn, oTrans, oLogger, oVer):
    """Copy AbstractCard, upgrading as needed"""
    # pylint: disable-msg=E1101, R0912
    # E1101 - SQLObject confuses pylint
    # R0912 - need the branches for this
    aMessages = []
    if oVer.check_tables_and_versions([AbstractCard],
            [AbstractCard.tableversion], oOrigConn):
        copy_abstract_card(oOrigConn, oTrans, oLogger)
    else:
        if oVer.check_tables_and_versions([AbstractCard], [3], oOrigConn):
            oACTable = AbstractCard_v3
            bHasBurn = True
        elif oVer.check_tables_and_versions([AbstractCard], [2], oOrigConn):
            aMessages.append('Missing data for the Burn Option on cards.'
                    ' You will need to reimport the White wolf card list'
                    ' for these to be correct')
            oACTable = AbstractCard_v2
            bHasBurn = False
        else:
            return (False, ["Unknown AbstractCard version"])
        for oCard in oACTable.select(connection=oOrigConn):
            oCardCopy = AbstractCard(id=oCard.id,
                    canonicalName=oCard.canonicalName, name=oCard.name,
                    text=oCard.text, connection=oTrans)
            oCardCopy.group = oCard.group
            oCardCopy.capacity = oCard.capacity
            oCardCopy.cost = oCard.cost
            oCardCopy.costtype = oCard.costtype
            oCardCopy.level = oCard.level
            oCardCopy.life = oCard.life
            if bHasBurn:
                oCardCopy.burnoption = oCard.burnoption
            else:
                oCardCopy.burnoption = False
            for oData in oCard.rarity:
                oCardCopy.addRarityPair(oData)
            for oData in oCard.discipline:
                oCardCopy.addDisciplinePair(oData)
            for oData in oCard.rulings:
                oCardCopy.addRuling(oData)
            for oData in oCard.clan:
                oCardCopy.addClan(oData)
            for oData in oCard.cardtype:
                oCardCopy.addCardType(oData)
            for oData in oCard.sect:
                oCardCopy.addSect(oData)
            for oData in oCard.title:
                oCardCopy.addTitle(oData)
            for oData in oCard.creed:
                oCardCopy.addCreed(oData)
            for oData in oCard.virtue:
                oCardCopy.addVirtue(oData)
            oCardCopy.syncUpdate()
            oLogger.info('copied AC %s', oCardCopy.name)
    return (True, aMessages)

def copy_physical_card(oOrigConn, oTrans, oLogger):
    """Copy PhysicalCard, assuming version match"""
    # We copy abstractCardID rather than abstractCard, to avoid issues
    # with abstract card class changes
    for oCard in PhysicalCard.select(connection=oOrigConn):
        oCardCopy = PhysicalCard(id=oCard.id,
                abstractCardID=oCard.abstractCardID,
                expansionID=oCard.expansionID, connection=oTrans)
        oLogger.info('copied PC %s', oCardCopy.id)

def get_new_physical_card(oCard, oTrans, aMessages):
    """Get the equivalent physical card in the new physical card table.

       If oCard is an AbstractCard, use None as the expansion.
       """
    if hasattr(oCard,'abstractCardID'):
        # must be a physical card
        oExp = oCard.expansion
        oId = oCard.abstractCardID
    else:
        oExp = None
        oId = oCard.id
    try:
        oNewCard = PhysicalCard.selectBy(
                    abstractCardID=oId,
                    expansion=oExp, connection=oTrans).getOne()
    except SQLObjectNotFound:
        if oExp is None:
            raise

        # probably an invalid (AbstractCard, Expansion) pair in the old
        # database. Try with no expansion.
        aMessages.append("'%s' isn't listed as appearing in" \
                         " expansion '%s'. Attempting to insert " \
                         " without an expansion." \
                         % (oCard.abstractCard.name, oCard.expansion.name))
        oNewCard = PhysicalCard.selectBy(
                    abstractCardID=oId,
                    expansion=None, connection=oTrans).getOne()
    return oNewCard

def copy_old_physical_card(oOrigConn, oTrans, oLogger, oVer):
    """Copy PhysicalCards, upgrading if needed."""
    aMessages = []
    if oVer.check_table_in_versions(PhysicalCardSet, [3, 4], oOrigConn) and \
            oVer.check_tables_and_versions([PhysicalCard],
                    [PhysicalCard.tableversion], oOrigConn):
        # We setup the list of cards and expansions as in the WW parser
        for oCard in AbstractCard.select(connection=oTrans):
            PhysicalCard(abstractCardID=oCard.id,
                    expansionID=None, connection=oTrans)
            for oExp in set([oRarity.expansion for oRarity in oCard.rarity]):
                PhysicalCard(abstractCardID=oCard.id,
                        expansionID=oExp.id, connection=oTrans)
            oTrans.commit()
            oLogger.info('created Physical Cards for AC %s', oCard.id)
        # We create a PhysicalCardSet named 'My Collection' and add the
        # changed cards here
        oPCLSet = PhysicalCardSet(name='My Collection', parent=None,
                connection=oTrans)
        # For all the Physical cards currently in the collection,
        # add them to the card set
        # pylint: disable-msg=E1101, E1103
        # SQLObject confuses pylint
        for oCard in PhysicalCard.select(connection=oOrigConn):
            oNewCard = get_new_physical_card(oCard, oTrans, aMessages)
            oPCLSet.addPhysicalCard(oNewCard.id)
            oLogger.info('copied PC %s', oNewCard.id)
    elif oVer.check_tables_and_versions([PhysicalCardSet, PhysicalCard],
            [PhysicalCardSet.tableversion, PhysicalCard.tableversion],
            oOrigConn):
        copy_physical_card(oOrigConn, oTrans, oLogger)
    else:
        return (False, ["Unknown PhysicalCard version"])
    return (True, aMessages)

def copy_physical_card_set(oOrigConn, oTrans, oLogger):
    """Copy PCS, assuming versions match"""
    # pylint: disable-msg=E1101
    # SQLObject confuses pylint
    aSets = list(PhysicalCardSet.select(connection=oOrigConn))
    bDone = False
    dDone = {}
    while not bDone:
        # We make sure we copy parent's before children
        # We need to be careful, since we don't retain card set IDs,
        # due to issues with sequence numbers
        aToDo = []
        for oSet in aSets:
            if oSet.parent is None or oSet.parent in dDone:
                if oSet.parent:
                    oParent = dDone[oSet.parent]
                else:
                    oParent = None
                oCopy = PhysicalCardSet(name=oSet.name,
                        author=oSet.author, comment=oSet.comment,
                        annotations=oSet.annotations, inuse=oSet.inuse,
                        parent=oParent, connection=oTrans)
                for oCard in oSet.cards:
                    oCopy.addPhysicalCard(oCard.id)
                oCopy.syncUpdate()
                oLogger.info('Copied PCS %s', oCopy.name)
                dDone[oSet] = oCopy
            else:
                aToDo.append(oSet)
        if not aToDo:
            bDone = True
        else:
            aSets = aToDo

def copy_old_physical_card_set(oOrigConn, oTrans, oLogger, oVer):
    """Copy PCS, upgrading as needed."""
    # pylint: disable-msg=E1101, E1103
    # SQLObject confuses pylint
    def _copy_cards(oOldSet, oNewSet):
        """Copy the cards from the old physical card set to the new set,
           replacing references to the correct new Physical Cards."""
        for oCard in oOldSet.cards:
            oNewCard = get_new_physical_card(oCard, oTrans, aMessages)
            oNewSet.addPhysicalCard(oNewCard.id)

    aMessages = []
    if oVer.check_tables_and_versions([PhysicalCardSet],
            [PhysicalCardSet.tableversion], oOrigConn) \
            and oVer.check_tables_and_versions([PhysicalCard],
                    [PhysicalCard.tableversion], oOrigConn):
        copy_physical_card_set(oOrigConn, oTrans, oLogger)
    else:
        if oVer.check_tables_and_versions([PhysicalCardSet], [3], oOrigConn):
            oPCSTable = PhysicalCardSet_v3
            bHasInUse = False
        elif oVer.check_tables_and_versions([PhysicalCardSet], [4], oOrigConn):
            oPCSTable = PhysicalCardSet_v4
            bHasInUse = True
        else:
            return (False, ["Unknown PhysicalCardSet version"])
        oParent = PhysicalCardSet.selectBy(name='My Collection',
                connection=oTrans).getOne()
        for oSet in oPCSTable.select(connection=oOrigConn):
            oCopy = PhysicalCardSet(name=oSet.name,
                    author=oSet.author, comment=oSet.comment,
                    annotations=None, inuse=False, parent=oParent,
                    connection=oTrans)
            _copy_cards(oSet, oCopy)
            if bHasInUse:
                oCopy.inuse = oSet.inuse
            oCopy.syncUpdate()
            oLogger.info('Copied PCS %s', oCopy.name)
    return (True, aMessages)

def copy_old_abstract_card_set(oOrigConn, oTrans, oLogger, oVer):
    """Copy old Abstract Card Sets, upgrading as needed."""
    # pylint: disable-msg=E1101
    # SQLObject confuses pylint
    def _gen_new_name(sSetName, aPhysSetNames):
        """Crate a new PCS name for the existing ACS"""
        sNewName = '(ACS) ' + sSetName
        sNewName = sNewName[:50]
        iSuffix = 0
        while sNewName in aPhysSetNames:
            sNewName = '(ACS) %s %d' % (sSetName[:40], iSuffix)
            sNewName = sNewName[:50]
            iSuffix += 1
        aPhysSetNames.append(sNewName)
        return sNewName

    def _copy_cards(oOldSet, oNewSet):
        """Copy the cards from the old abstract card set oOldSet to the
           PhysicalCardSet oNewSet, doing the right thing."""
        for oCard in oOldSet.cards:
            # We find the physical card with the same abstract card id
            # and No expansion info
            # pylint: disable-msg=E1103
            # SQLObject confuses pylint
            oNewCard = get_new_physical_card(oCard, oTrans, aMessages)
            oNewSet.addPhysicalCard(oNewCard.id)

    aMessages = []
    if oVer.check_tables_and_versions([AbstractCardSet_v3], [3], oOrigConn):
        aPhysSetNames = [x.name for x in
                PhysicalCardSet.select(connection=oTrans)]
        if oVer.check_tables_and_versions([AbstractCard], [3], oOrigConn):
            oACSTable = AbstractCardSet_v3
        elif oVer.check_tables_and_versions([AbstractCard], [2], oOrigConn):
            oACSTable = AbstractCardSet_ACv2
        else:
            return (False, ["Unknown AbstractCard version for the "
                "AbstractCardSet version"])
        for oSet in oACSTable.select(connection=oOrigConn):
            # Create a similiar Physical Card Set
            sNewName = _gen_new_name(oSet.name, aPhysSetNames)
            oNewPCS = PhysicalCardSet(name=sNewName, author=oSet.author,
                    annotations=oSet.annotations, comment=oSet.comment,
                    parent=None, inuse=False, connection=oTrans)
            _copy_cards(oSet, oNewPCS)
            oNewPCS.syncUpdate()
            oLogger.info('Copied Abstract CS %s', oNewPCS.name)
    else:
        return (False, ["Unknown AbstractCardSet version"])
    return (True, aMessages)

def read_old_database(oOrigConn, oDestConnn, oLogHandler=None):
    """Read the old database into new database, filling in
       blanks when needed"""
    try:
        if not check_can_read_old_database(oOrigConn):
            return False
    except UnknownVersion, oExp:
        raise oExp
    oLogger = Logger('read Old DB')
    if oLogHandler:
        oLogger.addHandler(oLogHandler)
        if hasattr(oLogHandler, 'set_total'):
            oLogHandler.set_total(old_database_count(oOrigConn))
    # OK, version checks pass, so we should be able to deal with this
    aMessages = []
    bRes = True
    oTrans = oDestConnn.transaction()
    # Magic happens in the individual functions, as needed
    aToCopy = [
            (copy_old_rarity, 'Rarity table', False),
            (copy_old_expansion, 'Expansion table', False),
            (copy_old_discipline, 'Discipline table', False),
            (copy_old_clan, 'Clan table', False),
            (copy_old_creed, 'Creed table', False),
            (copy_old_virtue, 'Virtue table', False),
            (copy_old_card_type, 'CardType table', False),
            (copy_old_ruling, 'Ruling table', False),
            (copy_old_discipline_pair, 'DisciplinePair table', False),
            (copy_old_rarity_pair, 'RarityPair table', False),
            (copy_old_sect, 'Sect table', False),
            (copy_old_title, 'Title table', False),
            (copy_old_abstract_card, 'AbstractCard table', True),
            (copy_old_physical_card, 'PhysicalCard table', True),
            (copy_old_physical_card_set, 'PhysicalCardSet table', True),
            (copy_old_abstract_card_set, 'AbstractCardSet table', True),
            ]
    oVer = DatabaseVersion()
    for fCopy, sName, bPassLogger in aToCopy:
        try:
            if bPassLogger:
                (bOK, aNewMessages) = fCopy(oOrigConn, oTrans, oLogger, oVer)
            else:
                (bOK, aNewMessages) = fCopy(oOrigConn, oTrans, oVer)
        except SQLObjectNotFound, oExp:
            bOK = False
            aNewMessages = ['Unable to copy %s: Error %s' % (sName, oExp)]
        else:
            if not bPassLogger:
                oLogger.info('%s copied' % sName)
        bRes = bRes and bOK
        aMessages.extend(aNewMessages)
        oTrans.commit()
        oTrans.cache.clear()
    return (bRes, aMessages)

# pylint: enable-msg=W0612, C0103

def drop_old_tables(oConn):
    """Drop tables whihc are no longer used from the database.
       Needed for postgres and other such things."""
    oVer = DatabaseVersion()
    if oVer.check_tables_and_versions([AbstractCardSet_v3], [3], oConn):
        # Need to drop the old ACS tables
        MapAbstractCardToAbstractCardSet_v3.dropTable(ifExists=True,
                connection=oConn)
        AbstractCardSet_v3.dropTable(ifExists=True, connection=oConn)

def copy_database(oOrigConn, oDestConnn, oLogHandler=None):
    """Copy the database, with no attempts to upgrade.

       This is a straight copy, with no provision for funky stuff
       Compatability of database structures is assumed, but not checked.
       """
    # Not checking versions probably should be fixed
    # Copy tables needed before we can copy AbstractCard
    flush_cache()
    oVer = DatabaseVersion()
    oVer.expire_cache()
    oLogger = Logger('copy DB')
    if oLogHandler:
        oLogger.addHandler(oLogHandler)
        if hasattr(oLogHandler, 'set_total'):
            iTotal = 12 + AbstractCard.select(connection=oOrigConn).count() + \
                    PhysicalCard.select(connection=oOrigConn).count() + \
                    PhysicalCardSet.select(connection=oOrigConn).count()
            oLogHandler.set_total(iTotal)
    bRes = True
    aMessages = []
    oTrans = oDestConnn.transaction()
    aToCopy = [
            (copy_rarity, 'Rarity table', False),
            (copy_expansion, 'Expansion table', False),
            (copy_discipline, 'Discipline table', False),
            (copy_clan, 'Clan table', False),
            (copy_creed, 'Creed table', False),
            (copy_virtue, 'Virtue table', False),
            (copy_card_type, 'CardType table', False),
            (copy_ruling, 'Ruling table', False),
            (copy_discipline_pair, 'DisciplinePair table', False),
            (copy_rarity_pair, 'RarityPair table', False),
            (copy_sect, 'Sect table', False),
            (copy_title, 'Title table', False),
            (copy_abstract_card, 'AbstractCard table', True),
            (copy_physical_card, 'PhysicalCard table', True),
            (copy_physical_card_set, 'PhysicalCardSet table', True),
            ]
    for fCopy, sName, bPassLogger in aToCopy:
        try:
            if bRes:
                if bPassLogger:
                    fCopy(oOrigConn, oTrans, oLogger)
                else:
                    fCopy(oOrigConn, oTrans)
        except SQLObjectNotFound, oExp:
            bRes = False
            aMessages.append('Unable to copy %s: Aborting with error: %s'
                    % (sName, oExp))
        else:
            oTrans.commit()
            oTrans.cache.clear()
            if not bPassLogger:
                oLogger.info('%s copied' % sName)
    flush_cache()
    # Clear out cache related joins and such
    return bRes, aMessages

def make_card_set_holder(oCardSet, oOrigConn):
    """Given a CardSet, create a Cached Card Set Holder for it."""
    oCurConn = sqlhub.processConnection
    sqlhub.processConnection = oOrigConn
    oCS = CachedCardSetHolder()
    oCS.name = oCardSet.name
    oCS.author = oCardSet.author
    oCS.comment = oCardSet.comment
    oCS.annotations = oCardSet.annotations
    oCS.inuse = oCardSet.inuse
    if oCardSet.parent:
        oCS.parent = oCardSet.parent.name
    for oCard in oCardSet.cards:
        if oCard.expansion is None:
            oCS.add(1, oCard.abstractCard.canonicalName, None)
        else:
            oCS.add(1, oCard.abstractCard.canonicalName, oCard.expansion.name)
    sqlhub.processConnection = oCurConn
    return oCS

def copy_to_new_abstract_card_db(oOrigConn, oNewConn, oCardLookup,
        oLogHandler=None):
    """Copy the card sets to a new Physical Card and Abstract Card List.

      Given an existing database, and a new database created from
      a new cardlist, copy the CardSets, going via CardSetHolders, so we
      can adapt to changed names, etc.
      """
    # pylint: disable-msg=R0914
    # we need a lot of variables here
    aPhysCardSets = []
    oOldConn = sqlhub.processConnection
    sqlhub.processConnection = oOrigConn
    # Copy Physical card sets
    oLogger = Logger('copy to new abstract card DB')
    if oLogHandler:
        oLogger.addHandler(oLogHandler)
        if hasattr(oLogHandler, 'set_total'):
            iTotal = 1 + PhysicalCardSet.select(connection=oOrigConn).count()
            oLogHandler.set_total(iTotal)
    aSets = list(PhysicalCardSet.select(connection=oOrigConn))
    bDone = False
    aDone = []
    # Ensre we only process a set after it's parent
    while not bDone:
        aToDo = []
        for oSet in aSets:
            if oSet.parent is None or oSet.parent in aDone:
                oCS = make_card_set_holder(oSet, oOrigConn)
                aPhysCardSets.append(oCS)
                aDone.append(oSet)
            else:
                aToDo.append(oSet)
        if not aToDo:
            bDone = True
        else:
            aSets = aToDo
    # Save the current mapping
    oLogger.info('Memory copies made')
    oTarget = oNewConn.transaction()
    sqlhub.processConnection = oTarget
    # Create the cardsets from the holders
    dLookupCache = {}
    for oSet in aPhysCardSets:
        oSet.create_pcs(oCardLookup, dLookupCache)
        oLogger.info('Physical Card Set: %s', oSet.name)
        oTarget.commit()
        oTarget.cache.clear()
    # Restore mapping
    oTarget = oNewConn.transaction()
    oTarget.commit()
    sqlhub.processConnection = oOldConn
    return (True, [])

def create_memory_copy(oTempConn, oLogHandler=None):
    """Create a temporary copy of the database in memory.

      We create a temporary memory database, and create the updated
      database in it. readOldDB is responsbile for upgrading stuff
      as needed
      """
    if refresh_tables(aObjectList, oTempConn):
        bRes, aMessages = read_old_database(sqlhub.processConnection,
                oTempConn, oLogHandler)
        oVer = DatabaseVersion()
        oVer.expire_cache()
        return bRes, aMessages
    return (False, ["Unable to create tables"])

def create_final_copy(oTempConn, oLogHandler=None):
    """Copy from the memory database to the real thing"""
    drop_old_tables(sqlhub.processConnection)
    if refresh_tables(aObjectList, sqlhub.processConnection):
        return copy_database(oTempConn, sqlhub.processConnection, oLogHandler)
    return (False, ["Unable to create tables"])

def attempt_database_upgrade(oLogHandler=None):
    """Attempt to upgrade the database, going via a temporary memory copy."""
    oTempConn = connectionForURI("sqlite:///:memory:")
    oLogger = Logger('attempt upgrade')
    if oLogHandler:
        oLogger.addHandler(oLogHandler)
    (bOK, aMessages) = create_memory_copy(oTempConn, oLogHandler)
    if bOK:
        oLogger.info("Copied database to memory, performing upgrade.")
        if len(aMessages) > 0:
            oLogger.info ("Messages reported: %s", aMessages)
        (bOK, aMessages) = create_final_copy(oTempConn, oLogHandler)
        if bOK:
            oLogger.info ("Everything seems to have gone OK")
            if len(aMessages) > 0:
                oLogger.info ("Messages reported %s", aMessages)
            return True
        else:
            oLogger.critical("Unable to perform upgrade.")
            if len(aMessages) > 0:
                oLogger.error("Errors reported: %s", aMessages)
            oLogger.critical("!!YOUR DATABASE MAY BE CORRUPTED!!")
    else:
        oLogger.error("Unable to create memory copy. Database not upgraded.")
        if len(aMessages) > 0:
            oLogger.error("Errors reported %s", aMessages)
    return False



# test_FilterParser.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2007, 2008 Simon Cross <hodgestar@gmail.com>,
# Copyright 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Tests the Filter Parser code"""

from sutekh.tests.TestCore import SutekhTest
from sutekh.tests.io import test_WhiteWolfParser
from sutekh.core.SutekhObjects import AbstractCard, PhysicalCardSet, \
        PhysicalCard, MapPhysicalCardToPhysicalCardSet, IAbstractCard, \
        IPhysicalCard, IExpansion
from sutekh.core import FilterParser, Filters
import unittest

class FilterParserTests(SutekhTest):
    """Class for the test cases"""
    aExpectedCards = test_WhiteWolfParser.WhiteWolfParserTests.aExpectedCards
    oFilterParser = FilterParser.FilterParser()

    # pylint: disable-msg=R0914
    # We don't really care about the number of local variables here
    def test_basic(self):
        """Set of simple tests of the filter parser."""
        # setup filters
        aTests = [
            # Single & Multiple value tests
            ('Clan in "Follower of Set"', [u"Aabbt Kindred", u"Abdelsobek",
                u"Amisa", u"Kemintiri (Advanced)"]),
            ('Clan in Ravnos, Samedi', [u"Abebe", u"L\xe1z\xe1r Dobrescu"]),
            ('Discipline in obf', [u"Aaron Bathurst", u"Abd al-Rashid",
                u"Abdelsobek", u"Abebe", u"Aeron", u"Amisa",
                u"Angelica, The Canonicus", u"Cedric", u"Kabede Maru",
                u"Kemintiri (Advanced)", u"Sha-Ennu"]),
            ('Discipline in fli', [u"Cedric"]),
            ('Discipline in nec, Quietus', [u"Abd al-Rashid", u"Abdelsobek",
                u"Abebe", u"Akram", u"Ambrogino Giovanni", u"Kabede Maru"]),
            ('Expansion_with_Rarity in Sabbat with Rare', [u"Ablative Skin"]),
            ('Expansion_with_Rarity in Third with Uncommon, Jyhad with Rare',
                [u"Aaron's Feeding Razor", u"Abbot"]),
            ('Discipline_with_Level in cel with superior', [u"Abd al-Rashid",
                u"Akram", u"Alexandra", u"Anson", u"Bronwen", u"Cesewayo",
                u"Kabede Maru"]),
            ('Discipline_with_Level in obt with inferior,'
                ' Potence with inferior, obf with superior',
                [u"Aaron Bathurst", u"Aaron Duggan, Cameron's Toady",
                    u"Aeron", u"Akram", u"Amisa", u"Bronwen", u"Cedric",
                    u"Kabede Maru", u"Kemintiri (Advanced)"]),
            ('CardType in Equipment', [u".44 Magnum", u"AK-47",
                u"Aaron's Feeding Razor"]),
            ('CardType in Reflex', [u"Predator's Communion"]),
            ('CardType in Power,Action', [u"Abbot", u"Abjure",
                u"Ablative Skin"]),
            ('Sect in Sabbat', [u"Aaron Bathurst",
                u"Aaron Duggan, Cameron's Toady", u"Aeron", u"Alfred Benezri",
                u"Angelica, The Canonicus", u"Bronwen", u"Sha-Ennu"]),
            ('Sect in Sabbat, Independent', [u"Aabbt Kindred",
                u"Aaron Bathurst", u"Aaron Duggan, Cameron's Toady",
                u"Abd al-Rashid", u"Abdelsobek", u"Abebe",
                u"Aeron", u"Alfred Benezri", u"Ambrogino Giovanni", u"Amisa",
                u"Angelica, The Canonicus", u"Bronwen",
                u"Kemintiri (Advanced)", u"L\xe1z\xe1r Dobrescu",
                u"Sha-Ennu"]),
            ('Title in Bishop', [u"Alfred Benezri"]),
            ('Title in "Independent with 1 vote"',
                [u"Ambrogino Giovanni"]),
            ('Title in Bishop, Prince',
                    [u"Alfred Benezri", u"Anson"]),
            ('Title in Magaji, Regent',
                    [u"Cesewayo", u"Kabede Maru", u"Sha-Ennu"]),
            ('Title in Primogen, Priscus, Cardinal',
                    [u"Akram", u"Angelica, The Canonicus", u"Bronwen"]),
            ('Creed = Martyr',
                    [u'Anna "Dictatrix11" Suljic']),
            ('Creed = Martyr, Innocent',
                    [u'Anna "Dictatrix11" Suljic',
                        u'Inez "Nurse216" Villagrande']),
            ('Virtue in Redemption', [u"Abjure",
                u'Anna "Dictatrix11" Suljic']),
            ('Group in 4', [u"Aaron Bathurst", u"Abebe",
                u'Anna "Dictatrix11" Suljic', u"Cedric", u"Cesewayo",
                u'Earl "Shaka74" Deams', u'Inez "Nurse216" Villagrande',
                u"Sha-Ennu"]),
            ('Group in 4, 5', [u"Aaron Bathurst", u"Abdelsobek", u"Abebe",
                u'Anna "Dictatrix11" Suljic', u"Cedric", u"Cesewayo",
                u'Earl "Shaka74" Deams', u'Inez "Nurse216" Villagrande',
                u"Kabede Maru", u"Sha-Ennu"]),
            ('Capacity in 2', [u"Aaron Duggan, Cameron's Toady"]),
            ('Capacity in 2,1', [u"Aaron Duggan, Cameron's Toady",
                u"Abombwe"]),
            ('Cost in 5', [u"AK-47"]),
            ('Cost in 2,5', [u".44 Magnum", u"AK-47"]),
            ('Cost in 0', [u'Aabbt Kindred', u'Aaron Bathurst',
                u"Aaron Duggan, Cameron's Toady", u'Abandoning the Flesh',
                u'Abbot', u'Abd al-Rashid', u'Abdelsobek', u'Abebe',
                u'Abjure', u'Ablative Skin', u'Abombwe', u'Aeron', u'Akram',
                u'Alan Sovereign', u'Alan Sovereign (Advanced)', u'Alexandra',
                u'Alfred Benezri', u'Ambrogino Giovanni', u'Amisa',
                u'Anastasz di Zagreb', u'Angelica, The Canonicus',
                u'Anna "Dictatrix11" Suljic', u'Anson', u'Bronwen',
                u'Cedric', u'Cesewayo', u'Earl "Shaka74" Deams',
                u'Gracis Nostinus', u'Inez "Nurse216" Villagrande',
                u'Kabede Maru', u'Kemintiri (Advanced)',
                u'L\xe1z\xe1r Dobrescu', u"Predator's Communion", u'Sha-Ennu',
                u'Yvette, The Hopeless']),
            ('Cost in 0,5', [u"AK-47", u'Aabbt Kindred', u'Aaron Bathurst',
                u"Aaron Duggan, Cameron's Toady", u'Abandoning the Flesh',
                u'Abbot', u'Abd al-Rashid', u'Abdelsobek', u'Abebe',
                u'Abjure', u'Ablative Skin', u'Abombwe', u'Aeron', u'Akram',
                u'Alan Sovereign', u'Alan Sovereign (Advanced)', u'Alexandra',
                u'Alfred Benezri', u'Ambrogino Giovanni', u'Amisa',
                u'Anastasz di Zagreb', u'Angelica, The Canonicus',
                u'Anna "Dictatrix11" Suljic', u'Anson', u'Bronwen',
                u'Cedric', u'Cesewayo', u'Earl "Shaka74" Deams',
                u'Gracis Nostinus', u'Inez "Nurse216" Villagrande',
                u'Kabede Maru', u'Kemintiri (Advanced)',
                u'L\xe1z\xe1r Dobrescu', u"Predator's Communion", u'Sha-Ennu',
                u'Yvette, The Hopeless']),
            ('Life in 6', [u'Anna "Dictatrix11" Suljic',
                u'Earl "Shaka74" Deams']),
            ('Life in 3, 6', [u'Anna "Dictatrix11" Suljic',
                u'Earl "Shaka74" Deams', u'Inez "Nurse216" Villagrande']),
            ('CostType in Pool', [u".44 Magnum", u"AK-47",
                u"Aaron's Feeding Razor", u"The Path of Blood"]),
            ('CostType in Blood', [u"Aire of Elation"]),
            ('CostType in Pool, Blood', [u".44 Magnum", u"AK-47",
                u"Aaron's Feeding Razor", u"Aire of Elation",
                u"The Path of Blood"]),
            ('Life in 4', []),
            ('Life in 4,5', []),

            # Text Entry Filters Filters
            ('CardText in strike', [u".44 Magnum", u"AK-47", u"Aeron",
                u"Anastasz di Zagreb", u"Bronwen"]),
            ('CardName in cameron', [u"Aaron Duggan, Cameron's Toady"]),
            (u'CardName in "L\xe1z\xe1r"', [u"L\xe1z\xe1r Dobrescu"]),

            # Compound Filters
            ('CardType in Equipment AND Cost in 5', [u"AK-47"]),
            ('CardType in Equipment OR CardType in Power',
                    [u".44 Magnum", u"AK-47", u"Aaron's Feeding Razor",
                        u"Abjure"]),
            ('NOT CardType in Equipment, Vampire',
                    [u"Abandoning the Flesh", u"Abbot", u"Abjure",
                        u"Ablative Skin", u"Abombwe", u"Aire of Elation",
                        u'Anna "Dictatrix11" Suljic', u'Earl "Shaka74" Deams',
                        u'Inez "Nurse216" Villagrande',
                        u"Predator's Communion", u"The Path of Blood"]),
            ('CardType not in Equipment, Vampire',
                    [u"Abandoning the Flesh", u"Abbot", u"Abjure",
                        u"Ablative Skin", u"Abombwe", u"Aire of Elation",
                        u'Anna "Dictatrix11" Suljic', u'Earl "Shaka74" Deams',
                        u'Inez "Nurse216" Villagrande',
                        u"Predator's Communion", u"The Path of Blood"]),
        ]

        # Abstract Card Filtering Tests
        for sFilter, aExpectedNames in aTests:
            oAST = self.oFilterParser.apply(sFilter)
            oFilter = oAST.get_filter()
            aCards = oFilter.select(AbstractCard).distinct()
            aNames = sorted([oC.name for oC in aCards])
            self.assertEqual(aNames, aExpectedNames,
                    "Filter Object %s failed."
                    " %s != %s." % (oFilter, aNames, aExpectedNames))

    def test_card_set_filters(self):
        """Tests for the physical card set filters."""
        # Although splitting this off does add an additional init
        # pass, the logical grouping is fairly different
        aCardSets = [('Test 1', 'Author A', 'A set', False),
                ('Test 2', 'Author B', 'Another set', False),
                ('Test 3', 'Author A', 'Something different', True)]
        aPCSCards = [
                # Set 1
                [('Abombwe', None), ('Alexandra', 'CE'),
                ('Sha-Ennu', None), ('Sha-Ennu', None), ('Sha-Ennu', None),
                ('Sha-Ennu', 'Third Edition')],
                # Set 2
                [('Sha-Ennu', 'Third Edition'), ('Anson', 'Jyhad'),
                    ('.44 magnum', 'Jyhad'), ('ak-47', 'LotN'),
                    ('Alexandra', 'CE'), ('Alexandra', 'CE')],
                # Set 3
                [('Yvette, The Hopeless', 'BSC')]]
        aPCSs = []
        # pylint: disable-msg=E1101
        # sqlobject confuses pylint
        for iCnt, tData in enumerate(aCardSets):
            sName, sAuthor, sComment, bInUse = tData
            oPCS = PhysicalCardSet(name=sName, comment=sComment,
                    author=sAuthor, inuse=bInUse)
            for sName, sExp in aPCSCards[iCnt]:
                if sExp:
                    oExp = IExpansion(sExp)
                else:
                    oExp = None
                oAbs = IAbstractCard(sName)
                oPhys = IPhysicalCard((oAbs, oExp))
                oPCS.addPhysicalCard(oPhys.id)
            aPCSs.append(oPCS)
        # Tests on the physical card set properties
        aPhysicalCardSetTests = [
                ('CardSetName = "Test 1"', [aPCSs[0]]),
                ('CardSetName = Test', sorted(aPCSs)),
                ('CSSetsInUse', [aPCSs[2]]),
                ('CardSetAuthor = "Author A"',
                    sorted([aPCSs[0], aPCSs[2]])),
                ('CardSetDescription = set',
                    sorted([aPCSs[0], aPCSs[1]])),
                ('CardSetDescription in different',
                    [aPCSs[2]]),
                ]

        for sFilter, aExpectedSets in aPhysicalCardSetTests:
            oAST = self.oFilterParser.apply(sFilter)
            oFilter = oAST.get_filter()
            aCardSets = sorted(oFilter.select(PhysicalCardSet).distinct())
            self.assertEqual(aCardSets, aExpectedSets, "Filter Object %s"
                    " failed. %s != %s." % (oFilter, aCardSets, aExpectedSets))

        # Test data for the Specific card filters

        aPCSAbsCardTests = [
                (Filters.PhysicalCardSetFilter('Test 1'),
                    'CardType in Vampire',
                    [u"Alexandra", u"Sha-Ennu", u"Sha-Ennu", u"Sha-Ennu",
                        u"Sha-Ennu"]),
                (Filters.PhysicalCardSetFilter('Test 1'),
                    'CardType in Master',
                    [u"Abombwe"]),
                (Filters.PhysicalCardSetFilter('Test 2'),
                    'CardType in Equipment',
                    [u".44 Magnum", u"AK-47"]),
                ]

        for oPCSFilter, sFilter, aExpectedCards in aPCSAbsCardTests:
            oAST = self.oFilterParser.apply(sFilter)
            oFilter = oAST.get_filter()
            oFullFilter = Filters.FilterAndBox([oPCSFilter, oFilter])
            aCSCards = [IAbstractCard(x).name for x in oFullFilter.select(
                        MapPhysicalCardToPhysicalCardSet).distinct()]
            self.assertEqual(aCSCards, aExpectedCards, "Filter Object %s"
                    " failed. %s != %s." % (oFullFilter, aCSCards,
                        aExpectedCards))

        oInUseFilter = self.oFilterParser.apply('SetsInUse').get_filter()
        aPCSCardsInUse = list(oInUseFilter.select(PhysicalCard).distinct())
        aExpectedCards = list(aPCSs[2].cards)
        self.assertEqual(aPCSCardsInUse, aExpectedCards, 'PhysicalCardSet In '
                'Use Filter failed %s != %s' % (aPCSCardsInUse,
                    aExpectedCards))

        oInUseFilter = self.oFilterParser.apply('CSSetsInUse').get_filter()
        aCSInUse = list(oInUseFilter.select(PhysicalCardSet).distinct())
        aExpectedSets = [aPCSs[2]]
        self.assertEqual(aCSInUse, aExpectedSets, 'CSPhysicalCardSet In'
                'Use Filter failed %s != %s' % (aCSInUse, aExpectedSets))

        # Number tests
        aPCSNumberTests = [
                (Filters.PhysicalCardSetFilter('Test 1'),
                    'CardCount = 4 from "%s"' % aPCSs[0].name,
                    [u"Sha-Ennu", u"Sha-Ennu", u"Sha-Ennu", u"Sha-Ennu"]),
                (Filters.PhysicalCardSetFilter('Test 1'),
                    'CardCount = 7 from "%s"' % aPCSs[0].name,
                    []),
                (Filters.PhysicalCardSetFilter('Test 1'),
                    'CardCount in "1" from "%s"' % aPCSs[0].name,
                    [u"Abombwe", u"Alexandra"]),
                (Filters.PhysicalCardSetFilter('Test 1'),
                    'CardCount in 1, 4 from "%s"' % aPCSs[0].name,
                    [u"Abombwe", u"Alexandra", u"Sha-Ennu", u"Sha-Ennu",
                        u"Sha-Ennu", u"Sha-Ennu"]),
                # Cards in 'Test 2' with zero count in 'Test 1'
                (Filters.PhysicalCardSetFilter('Test 2'),
                    'CardCount = 0 from "%s"' % aPCSs[0].name,
                    [u"Anson", u".44 Magnum", u"AK-47"]),
                (Filters.PhysicalCardSetFilter('Test 2'),
                    'CardCount = 0 from "%s" and CardType = Vampire' %
                    aPCSs[0].name,
                    [u"Anson"]),
                    ]

        for oPCSFilter, sFilter, aExpectedCards in aPCSNumberTests:
            # pylint: disable-msg=E1101
            # pyprotocols confuses pylinta
            oAST = self.oFilterParser.apply(sFilter)
            oFilter = oAST.get_filter()
            oFullFilter = Filters.FilterAndBox([oPCSFilter, oFilter])
            aCSCards = [IAbstractCard(x).name for x in oFullFilter.select(
                        MapPhysicalCardToPhysicalCardSet).distinct()]
            self.assertEqual(aCSCards, aExpectedCards, "Filter Object %s"
                    " failed. %s != %s." % (oFullFilter, aCSCards,
                        aExpectedCards))

    def test_variables(self):
        """Test parsing filters with variables"""

        aTests = ['CardType = $a', 'CardType in $a', 'Clan in $a']

        for sFilter in aTests:
            oAST = self.oFilterParser.apply(sFilter)
            oFilterPart = oAST.aChildren[0] # Because they're simple children
            self.assertEqual(oFilterPart.get_name(), '$a')
            self.assertTrue(oFilterPart.aFilterValues is None)

        # Test error case
        self.assertRaises(ValueError, self.oFilterParser.apply,
            'CardType in $a and Clan in $a')
        self.assertRaises(ValueError, self.oFilterParser.apply,
            'CardType in')
        self.assertRaises(ValueError, self.oFilterParser.apply,
            'Vampire in $a')

        oAST = self.oFilterParser.apply('CardType in "Vampire","Action Mod"')
        self.assertEqual(oAST.get_invalid_values(), ["Action Mod"])


if __name__ == "__main__":
    unittest.main()



# test_IdentifyXMLFile.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Test IdentifyXMLFile handling"""

from sutekh.tests.TestCore import SutekhTest
from sutekh.io.IdentifyXMLFile import IdentifyXMLFile
from sutekh.core.SutekhObjects import IAbstractCard, IPhysicalCard
import unittest

class TestIdentifyXMLFile(SutekhTest):
    """class for the IdentifyXMLFile tests"""

    def test_identify_xml_file(self):
        """Test IdentifyXMLFile"""
        # pylint: disable-msg=E1101
        # E1101: SQLObject + PyProtocols magic confuses pylint
        aAbstractCards = ['.44 magnum', 'ak-47', 'abbot', 'abebe', 'abombwe']
        # test IO

        oAC = IAbstractCard(aAbstractCards[0])
        oPC = IPhysicalCard((oAC, None))
        sLastWriterVersion = "1.0"

        sExample = '<cards sutekh_xml_version="%s"><card count="1" ' \
                   'expansion="None Specified" id="%d" name="%s" /></cards>' \
                   % (sLastWriterVersion, oPC.id, oAC.name)

        sExampleACS = '<abstractcardset author="A test author" ' \
                'comment="A test comment" name="Test Set 1" ' \
                'sutekh_xml_version="1.1"><annotations /><card count="1" ' \
                'id="11" name="Abebe" /><card count="1" id="8" name="Abbot" ' \
                '/><card count="1" id="1" name=".44 Magnum" /><card ' \
                'count="1" id="2" name="AK-47" /><card count="1" id="14" ' \
                'name="Abombwe" /></abstractcardset>'

        sExamplePCS = '<physicalcardset author="A test author" ' \
                'comment="A test comment" name="Test Set 1" '\
                'sutekh_xml_version="1.2"><annotations /><card count="1" ' \
                'expansion="None Specified" id="11" name="Abebe" />' \
                '</physicalcardset>'

        sTempFileName =  self._create_tmp_file()
        fOut = open(sTempFileName, 'w')
        fOut.write(sExample)
        fOut.close()

        oIdFile = IdentifyXMLFile()
        oIdFile.id_file(sTempFileName)
        self.assertEqual(oIdFile.type, 'PhysicalCard')

        oIdFile.parse_string(sExample)
        self.assertEqual(oIdFile.type, 'PhysicalCard')

        oIdFile.parse_string('garbage input')
        self.assertEqual(oIdFile.type, 'Unknown')

        oIdFile.parse_string(sExampleACS)
        self.assertEqual(oIdFile.type, 'AbstractCardSet')

        oIdFile.parse_string(sExamplePCS)
        self.assertEqual(oIdFile.type, 'PhysicalCardSet')

if __name__ == "__main__":
    unittest.main()



# CardListModel.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2006 Simon Cross <hodgestar@gmail.com>
# Copyright 2006 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""The gtk.TreeModel for the card lists."""

import gtk, pango, gobject
from sutekh.core.Filters import FilterAndBox, NullFilter, PhysicalCardFilter
from sutekh.core.Groupings import CardTypeGrouping
from sutekh.core.SutekhObjects import IAbstractCard, PhysicalCard, \
        IPhysicalCard

def remove_markup(sMarkup):
    """Return the string with markup stripped"""
    # Wrapper around parse_markup, dropping the info we wish to ignore
    return pango.parse_markup(sMarkup)[1]

def norm_path(oPath):
    """Transform string paths to tuple paths"""
    # Some widgets give us a path string, others a tuple,
    # to deal with tuples when moving between expansions and
    # card names
    if isinstance(oPath, str):
        oNormPath = tuple([int(x) for x in oPath.split(':')])
    else:
        oNormPath = oPath
    return oNormPath


class CardListModelListener(object):
    """Listens to updates, i.e. .load(...), .alter_card_count(...),
       .add_new_card(..) calls, to CardListModels."""
    def load(self, aAbsCards):
        """The CardListModel has reloaded itself. aAbsCards is the list of
           AbstractCards loaded."""
        pass

    def alter_card_count(self, oCard, iChg):
        """The count of the given card has been altered by iChg.

           oCard: AbstractCard for the card altered (the actual card may be
           a Physical Card).
           """
        pass

    def add_new_card(self, oCard):
        """A single copy of the given card has been added to this set.

           oCard: AbstractCard for the card altered (the actual card may be
           a Physical Card).
           """
        pass

class CardListModel(gtk.TreeStore):
    # pylint: disable-msg=R0904, R0902
    # inherit a lot of public methods for gtk, need local attributes for state
    """Provides a card list specific API for accessing a gtk.TreeStore."""
    # Use spaces to ensure it sorts first
    # Could possibly be more visually distinct, but users can filter
    # on unknown expansions if needed.

    sUnknownExpansion = '  Unspecified Expansion'

    def __init__(self):
        # STRING is the card name, INT is the card count
        super(CardListModel, self).__init__(str, str, str, bool, bool,
                gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT)
        # name, count, parent count, showInc, showDec, text_list, icons
        self._dName2Iter = {}
        self._dNameExpansion2Iter = {}
        self._dGroupName2Iter = {}

        self._cGroupBy = CardTypeGrouping # grouping class to use
        # base filter defines the card list
        self._oBaseFilter = PhysicalCardFilter()
        self._cCardClass = PhysicalCard # card class to use
        self._bApplyFilter = False # whether to apply the select filter
        # additional filters for selecting from the list
        self._oSelectFilter = None

        self.dListeners = {} # dictionary of CardListModelListeners

        self.bExpansions = True
        self.oEmptyIter = None
        self.oIconManager = None
        self.bUseIcons = True

    # pylint: disable-msg=W0212, C0103
    # W0212 - we explicitly allow access via these properties
    # C0103 - we allow these names
    cardclass = property(fget=lambda self: self._cCardClass,
            fset=lambda self, x: setattr(self, '_cCardClass', x))
    groupby = property(fget=lambda self: self._cGroupBy,
            fset=lambda self, x: setattr(self, '_cGroupBy', x))
    basefilter = property(fget=lambda self: self._oBaseFilter,
            fset=lambda self, x: setattr(self, '_oBaseFilter', x))
    applyfilter = property(fget=lambda self: self._bApplyFilter,
            fset=lambda self, x: setattr(self, '_bApplyFilter', x))
    selectfilter = property(fget=lambda self: self._oSelectFilter,
            fset=lambda self, x: setattr(self, '_oSelectFilter', x))
    # pylint: enable-msg=W0212, C0103

    def add_listener(self, oListener):
        """Add a listener to the list of interested listeners."""
        self.dListeners[oListener] = None

    def remove_listener(self, oListener):
        """Remove a listener from the list."""
        del self.dListeners[oListener]

    # various utilty functions for checking the model state

    # pylint: disable-msg=W0613
    # oCard needed so children can override this sensibly
    def get_expansion_info(self, oCard, dExpanInfo):
        """Get information about expansions"""
        aExpansions = []
        if not self.bExpansions:
            return aExpansions
        for oExpansion in dExpanInfo:
            aExpansions.append(self.get_expansion_name(oExpansion))
        return aExpansions

    # pylint: enable-msg=W0613

    def get_expansion_name(self, oExpansion):
        """Utility function to return iether the name, or the appropriate
           placeholder for oExpansion is None."""
        if oExpansion:
            return oExpansion.name
        return self.sUnknownExpansion

    def get_int_value(self, oIter, iPos):
        """Extract an integer value from the model, removing markup"""
        sMarkup = self.get_value(oIter, iPos)
        if sMarkup:
            sValue = remove_markup(sMarkup)
        else:
            return 0
        return int(sValue)

    def num_col_sort_func(self, oModel, oIter1, oIter2, iPos):
        """Custom sort function for the numerical columns

           iPos is user info set by set_sort_func"""
        iNum1 = self.get_int_value(oIter1, iPos)
        iNum2 = self.get_int_value(oIter2, iPos)
        return cmp(iNum1, iNum2)

    def lookup_icons(self, sGroup):
        """Lookup the icons for the group. Method since it's repeated in
           several places in CardSetListModel"""
        if self.oIconManager and self.bUseIcons:
            aTexts, aIcons = self.oIconManager.get_info(sGroup,
                    self.groupby)
        else:
            aTexts = aIcons = []
        return aTexts, aIcons

    def load(self):
        # pylint: disable-msg=R0914
        # we use many local variables for clarity
        """Clear and reload the underlying store. For use after initialisation
           or when the filter or grouping changes."""
        self.clear()
        self._dName2Iter = {}
        self._dNameExpansion2Iter = {}
        self._dGroupName2Iter = {}

        oCardIter = self.get_card_iterator(self.get_current_filter())
        # pylint: disable-msg=W0612
        # fGetCount is not used here
        fGetCard, fGetCount, fGetExpanInfo, oGroupedIter, aAbsCards = \
                self.grouped_card_iter(oCardIter)

        self.oEmptyIter = None

        # Iterate over groups
        for sGroup, oGroupIter in oGroupedIter:
            # Check for null group
            if sGroup is None:
                sGroup = '<< None >>'

            # Create Group Section
            oSectionIter = self.append(None)
            self._dGroupName2Iter[sGroup] = oSectionIter

            # Fill in Cards
            for oItem in oGroupIter:
                oCard = fGetCard(oItem)
                oChildIter = self.append(oSectionIter)
                self.set(oChildIter,
                    0, oCard.name,
                    5, [],
                    6, [],
                )
                aExpansionInfo = self.get_expansion_info(oCard,
                        fGetExpanInfo(oItem))
                for sExpansion in sorted(aExpansionInfo):
                    oExpansionIter = self.append(oChildIter)
                    self.set(oExpansionIter,
                            0, sExpansion,
                            5, [],
                            6, [])
                    self._dNameExpansion2Iter.setdefault(oCard.name,
                            {}).setdefault(sExpansion, []).append(
                                    oExpansionIter)
                self._dName2Iter.setdefault(oCard.name, []).append(oChildIter)

            # Update Group Section
            aTexts, aIcons = self.lookup_icons(sGroup)
            if aTexts:
                self.set(oSectionIter,
                    0, sGroup,
                    5, aTexts,
                    6, aIcons,
                )
            else:
                self.set(oSectionIter,
                    0, sGroup,
                    5, [],
                    6, [],
                )

        if not self._dName2Iter:
            # Showing nothing
            self.oEmptyIter = self.append(None)
            sText = self._get_empty_text()
            self.set(self.oEmptyIter, 0, sText, 5, [], 6, [])

        # Notify Listeners
        for oListener in self.dListeners:
            oListener.load(aAbsCards)

    def get_card_iterator(self, oFilter):
        """Return an interator over the card model.

           The filter is combined with self.basefilter. None may be used to
           retrieve the entire card list (with only the base filter
           restriciting which cards appear).
           """
        oFilter = self.combine_filter_with_base(oFilter)

        return oFilter.select(self.cardclass).distinct()

    def grouped_card_iter(self, oCardIter):
        """Return iterator over the card list grouping.

           Returns a triple of fGetCard (the function used to
           retrieve a card from an item), fGetCount (the function used to
           retrieve a card count from an item) and oGroupedIter (an iterator
           over the card groups)
           """
        aAbsCards = []
        fGetCard = lambda x:x[0]
        fGetCount = lambda x:x[1][0]
        fGetExpanInfo = lambda x:x[1][1]

        # Count by Abstract Card
        dAbsCards = {}

        for oCard in oCardIter:
            # pylint: disable-msg=E1101
            # sqlobject confuses pylint
            oPhysCard = IPhysicalCard(oCard)
            oAbsCard = IAbstractCard(oPhysCard)
            aAbsCards.append(oAbsCard)
            dAbsCards.setdefault(oAbsCard, [0, {}])
            dAbsCards[oAbsCard][0] += 1
            if self.bExpansions:
                dExpanInfo = dAbsCards[oAbsCard][1]
                dExpanInfo.setdefault(oPhysCard.expansion, 0)
                dExpanInfo[oPhysCard.expansion] += 1

        aCards = list(dAbsCards.iteritems())
        aCards.sort(lambda x, y: cmp(x[0].name, y[0].name))

        # Iterate over groups
        return (fGetCard, fGetCount, fGetExpanInfo,
                self.groupby(aCards, fGetCard), aAbsCards)

    def get_current_filter(self):
        """Get the current applied filter."""
        if self.applyfilter:
            return self.selectfilter
        else:
            return None

    def combine_filter_with_base(self, oOtherFilter):
        """Return the combination of oOtherFilter with the base filter.

           This handles the cases where either filter is None properly."""
        if self.basefilter is None and oOtherFilter is None:
            return NullFilter()
        elif self.basefilter is None:
            return oOtherFilter
        elif oOtherFilter is None:
            return self.basefilter
        else:
            return FilterAndBox([self.basefilter, oOtherFilter])

    def get_card_name_from_path(self, oPath):
        """Get the card name associated with the current path. Handle the
           expansion level transparently."""
        oIter = self.get_iter(oPath)
        if self.iter_depth(oIter) > 1:
            # Child of the card name
            # According to the docs, this will be the correct path
            oIter = self.get_iter(norm_path(oPath)[0:2])
        return self.get_name_from_iter(oIter)

    def get_all_from_path(self, oPath):
        """Get all relevent information about the current path.

           Returns the tuple (CardName, Expansion info, Card Count,
           depth in the  model), where depth in the model is 1 for the top
           level of cards, and 2 for the expansion level.
           """
        oIter = self.get_iter(oPath)
        iDepth = self.iter_depth(oIter)
        if iDepth == 2:
            sName = self.get_name_from_iter(self.get_iter(
                norm_path(oPath)[0:2]))
            sExpansion = self.get_value(oIter, 0)
        else:
            sName = self.get_name_from_iter(oIter)
            sExpansion = None
        iCount = self.get_int_value(oIter, 1)
        return sName, sExpansion, iCount, iDepth

    def get_child_entries_from_path(self, oPath):
        """Return a list of (sExpansion, iCount) pairs for the children of
           this path"""
        oIter = self.get_iter(oPath)
        aChildren = []
        iDepth = self.iter_depth(oIter)
        if iDepth != 1:
            # No children to look at
            return aChildren
        oChildIter = self.iter_children(oIter)
        while oChildIter:
            sExpansion = self.get_value(oChildIter, 0)
            iCount = self.get_int_value(oChildIter, 1)
            aChildren.append((sExpansion, iCount))
            oChildIter = self.iter_next(oChildIter)
        return aChildren

    def get_inc_dec_flags_from_path(self, oPath):
        """Get the settings of the inc + dec flags for the current path"""
        oIter = self.get_iter(oPath)
        bInc = self.get_value(oIter, 3)
        bDec = self.get_value(oIter, 4)
        return (bInc, bDec)

    def get_exp_name_from_path(self, oPath):
        """Get the expansion information from the model, returing None if this
           is not at a level where the expansion is known."""
        oIter = self.get_iter(oPath)
        if self.iter_depth(oIter) != 2:
            return None
        return self.get_name_from_iter(oIter)

    def get_name_from_iter(self, oIter):
        """Extract the value at oIter from the model, correcting for encoding
           issues."""
        # For some reason the string comes back from the
        # tree store having been encoded *again* despite
        # displaying correctly, so we decode it here.
        # I hope all systems encode with utf-8. :(
        sCardName = self.get_value(oIter, 0).decode("utf-8")
        return sCardName

    def get_card_count_from_iter(self, oIter):
        """Return the card count for a given iterator"""
        return self.get_int_value(oIter, 1)

    def get_parent_count_from_iter(self, oIter):
        """Return the parent count for a given iterator"""
        return self.get_int_value(oIter, 2)

    def _get_empty_text(self):
        """Get the correct text for an empty model."""
        if self.get_card_iterator(None).count() == 0:
            sText = 'Empty'
        else:
            sText = 'No Cards found'
        return sText



# CardSetMenu.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Menu for the CardSet View's
# Copyright 2006 Neil Muller <drnlmuller+sutekh@gmail.com>
# Copyright 2006 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details

"""Menu for the card set pane"""

import gtk
from sutekh.SutekhUtility import safe_filename
from sutekh.core.SutekhObjects import PhysicalCardSet
from sutekh.gui.SutekhFileWidget import ExportDialog
from sutekh.io.XmlFileHandling import PhysicalCardSetXmlFile
from sutekh.gui.PaneMenu import CardListMenu
from sutekh.gui.CardSetListModel import NO_SECOND_LEVEL, SHOW_EXPANSIONS, \
        SHOW_CARD_SETS, EXPANSIONS_AND_CARD_SETS, CARD_SETS_AND_EXPANSIONS, \
        THIS_SET_ONLY, ALL_CARDS, PARENT_CARDS, CHILD_CARDS, \
        IGNORE_PARENT, PARENT_COUNT, MINUS_THIS_SET, MINUS_SETS_IN_USE

class CardSetMenu(CardListMenu):
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    """Card Set Menu.

       Provide the usual menu options, and implement several of the
       card set specific actions - editing card set properties, editng
       annotations, exporting to file, and so on.
       """
    # pylint: disable-msg=R0902
    # R0902 - we are keeping a lot of state, so many instance variables
    def __init__(self, oFrame, oController, oWindow, sName):
        super(CardSetMenu, self).__init__(oFrame, oWindow, oController)
        self.__oController = oController
        self.sSetName = sName
        self.__create_card_set_menu()
        self.create_edit_menu()
        self.create_filter_menu()
        self.create_plugins_menu('_Plugins', self._oFrame)

    # pylint: disable-msg=W0201
    # these methods are called from __init__, so it's OK
    def __create_card_set_menu(self):
        """Create the Actions menu for Card Sets."""
        oMenu = self.create_submenu(self, '_Actions')
        self.create_menu_item("Edit Card Set _Properties", oMenu,
                self._edit_properties)
        self.create_menu_item("_Save Card Set to File", oMenu,
                self._do_export)
        # Submenu for plugins
        self.create_submenu(oMenu, "_Export Card Set")
        oMenu.add(gtk.SeparatorMenuItem())
        oCountMenu = self.create_menu_item_with_submenu(oMenu,
                "Cards To Show").get_submenu()
        oThisCardSet = None
        for sString, iValue in [
                ("This Card Set Only", THIS_SET_ONLY),
                ("Show All Cards", ALL_CARDS),
                ("Show all cards in parent card set", PARENT_CARDS),
                ("Show all cards in child card sets", CHILD_CARDS),
                ]:
            oItem = gtk.RadioMenuItem(oThisCardSet, sString)
            if not oThisCardSet:
                oThisCardSet = oItem
            if self.__oController.model.iShowCardMode == iValue:
                oItem.set_active(True)
            oCountMenu.add(oItem)
            oItem.connect("toggled", self._change_count_mode, iValue)
        oModeMenu = self.create_menu_item_with_submenu(oMenu,
                "Display Mode").get_submenu()
        oNoChildren = None
        for sString, iValue in [
                ("Show No Children", NO_SECOND_LEVEL),
                ("Show Expansions", SHOW_EXPANSIONS),
                ("Show Child Card Sets", SHOW_CARD_SETS),
                ("Show Expansions and Child Card Sets",
                    EXPANSIONS_AND_CARD_SETS),
                ("Show Child Card Sets and Expansions",
                    CARD_SETS_AND_EXPANSIONS),
                ]:
            oItem = gtk.RadioMenuItem(oNoChildren, sString)
            if not oNoChildren:
                oNoChildren = oItem
            if self.__oController.model.iExtraLevelsMode == iValue:
                oItem.set_active(True)
            oModeMenu.add(oItem)
            oItem.connect("toggled", self._change_mode, iValue)

        self._oParentCol = self.create_menu_item_with_submenu(oMenu,
                "Parent Card Count")
        oParentCountMenu = self._oParentCol.get_submenu()
        oNoParentCount = None
        # pylint: disable-msg=E1101
        # SQLObject confuses pylint
        oCS = PhysicalCardSet.byName(self.sSetName)
        for sString, iValue in [
                ("Don't show parent card counts", IGNORE_PARENT),
                ("Show Parent Count", PARENT_COUNT),
                ("Show difference between Parent Count and this card set",
                    MINUS_THIS_SET),
                ("Show difference between Parent Count and card sets in use",
                    MINUS_SETS_IN_USE),
                ]:
            # Should we have some way of restoring setting from last session?
            oItem = gtk.RadioMenuItem(oNoParentCount, sString)
            if not oNoParentCount:
                oNoParentCount = oItem
            if iValue == PARENT_COUNT:
                self._oDefaultParentCount = oItem
            if self.__oController.model.iParentCountMode == iValue and \
                    oCS.parent:
                oItem.set_active(True)
                self._change_parent_count_mode(oItem, iValue, True)
            elif not oCS.parent and iValue == IGNORE_PARENT:
                self._oParentCol.set_sensitive(False)
                oItem.set_active(True)
                self._change_parent_count_mode(oItem, iValue, True)
            oParentCountMenu.add(oItem)
            oItem.connect("toggled", self._change_parent_count_mode, iValue,
                    False)

        self.create_check_menu_item('Show icons for the grouping',
                oMenu, self.__oController.toggle_icons, True)

        oMenu.add(gtk.SeparatorMenuItem())
        self.add_common_actions(oMenu)

        oMenu.add(gtk.SeparatorMenuItem())
        self.create_menu_item("Delete Card Set", oMenu, self._card_set_delete)

    def create_edit_menu(self):
        """Create the 'Edit' menu, and populate it."""
        oMenu = self.create_submenu(self, "_Edit")
        self._oEditable = self.create_check_menu_item('Card Set is Editable',
                oMenu, self.toggle_editable, False, '<Ctrl>e')
        self._oController.view.set_menu(self)
        self.create_menu_item('Copy selection', oMenu, self.copy_selection,
                '<Ctrl>c')
        self._oPaste = self.create_menu_item('Paste', oMenu,
                self._paste_selection, '<Ctrl>v')
        self._oDel = self.create_menu_item('Delete selection', oMenu,
                self._del_selection, 'Delete')
        self._oPaste.set_sensitive(False)
        self._oDel.set_sensitive(False)
        self.add_edit_menu_actions(oMenu)

    # pylint: enable-msg=W0201

    # pylint: disable-msg=W0613
    # oWidget required by function signature for the following methods
    def _edit_properties(self, oWidget):
        """Popup the Edit Properties dialog to change card set properties."""
        # pylint: disable-msg=E1101
        # sqlobject confuses pylint
        self._oController.edit_properties(self)

    def update_card_set_menu(self, oCardSet):
        """Update the menu to reflect changes in the card set name."""
        self.sSetName = oCardSet.name
        self._oFrame.update_name(self.sSetName)

    def check_parent_count_column(self, oOldParent, oNewParent):
        """Check that the parent column values are set correctly
           when the parent changes.
           """
        # pylint: disable-msg=E1101
        # SQLObject confuses pylint
        # We rely on signal handler to cause reload
        if oNewParent:
            self._oParentCol.set_sensitive(True)
            if not oOldParent:
                # Parent has changed from none, so set to default
                self._change_parent_count_mode(None, PARENT_COUNT, True)
                # Check below will force reload anyway
                self._oDefaultParentCount.set_active(True)
        else:
            self._oParentCol.set_sensitive(False)
            self._change_parent_count_mode(None, IGNORE_PARENT, True)

    def _do_export(self, oWidget):
        """Export the card set to the chosen filename."""
        oFileChooser = ExportDialog("Save Card Set As ", self._oMainWindow,
                '%s.xml' % safe_filename(self.sSetName))
        oFileChooser.add_filter_with_pattern('XML Files', ['*.xml'])
        oFileChooser.run()
        sFileName = oFileChooser.get_name()
        if sFileName is not None:
            # User has OK'd us overwriting anything
            oWriter = PhysicalCardSetXmlFile(sFileName)
            oWriter.write(self.sSetName)

    def _card_set_delete(self, oWidget):
        """Delete the card set."""
        self._oController.delete_card_set()

    def _change_mode(self, oWidget, iLevel):
        """Set which extra information is shown."""
        # NB. gtk calls this twice, once for the current value,
        # and once for the new value. We don't want to call
        # load multiple times, so this logic
        if self._oController.model.iExtraLevelsMode != iLevel:
            self._oController.model.iExtraLevelsMode = iLevel
            self._oController.view.reload_keep_expanded()

    def _change_count_mode(self, oWidget, iLevel):
        """Set which extra information is shown."""
        if self._oController.model.iShowCardMode != iLevel:
            self._oController.model.iShowCardMode = iLevel
            self._oController.view.reload_keep_expanded()

    def _change_parent_count_mode(self, oWidget, iLevel, bNoReload=False):
        """Toggle the visibility of the parent col"""
        if iLevel == IGNORE_PARENT:
            self._oController.view.set_parent_count_col_vis(False)
        else:
            self._oController.view.set_parent_count_col_vis(True)
        if self._oController.model.iParentCountMode == iLevel:
            return
        self._oController.model.iParentCountMode = iLevel
        if not bNoReload:
            self._oController.view.reload_keep_expanded()

    def toggle_editable(self, oWidget):
        """Toggle the editable state of the card set."""
        if self._oController.model.bEditable != oWidget.active:
            self._oController.view.toggle_editable(oWidget.active)
            self._oPaste.set_sensitive(oWidget.active)
            self._oDel.set_sensitive(oWidget.active)

    def force_editable_mode(self, bValue):
        """Toggle the editable state of the card set."""
        self._oEditable.set_active(bValue)
        self._oPaste.set_sensitive(bValue)
        self._oDel.set_sensitive(bValue)

    def _del_selection(self, oWidget):
        """Delete the current selection"""
        self._oController.view.del_selection()

    def _paste_selection(self, oWidget):
        """Try to paste the current clipbaord contents"""
        self._oController.view.do_paste()

    # pylint: enable-msg=W0613



# CardImages.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Adds a frame which will display card images from ARDB in the GUI"""

import gtk
import gobject
import unicodedata
import os
import zipfile
import urllib2
import tempfile
from sqlobject import SQLObjectNotFound
from sutekh.core.SutekhObjects import PhysicalCard, PhysicalCardSet, \
        IAbstractCard, IExpansion
from sutekh.gui.PluginManager import CardListPlugin
from sutekh.gui.ProgressDialog import ProgressDialog
from sutekh.gui.CardListView import CardListViewListener
from sutekh.gui.BasicFrame import BasicFrame
from sutekh.gui.SutekhDialog import SutekhDialog, do_complaint_buttons, \
        do_complaint_error
from sutekh.gui.AutoScrolledWindow import AutoScrolledWindow
from sutekh.SutekhUtility import prefs_dir, ensure_dir_exists


_iFORWARD, _iBACKWARD = range(2)
_iFULL, _iVIEW_FIXED, _iFIT = range(3)
_tRATIO = (225, 300)

def _scale_dims(iImageWidth, iImageHeight, iPaneWidth, iPaneHeight):
    """Rescale the image dimension so they fit in the pane, preserving the
       aspect ratiom."""
    fImageAspectRatio = float(iImageHeight) / float(iImageWidth)
    fPaneAspectRatio = float(iPaneHeight) / float(iPaneWidth)

    if fPaneAspectRatio > fImageAspectRatio:
        # wider
        fDestWidth = iPaneWidth
        fDestHeight = iPaneWidth * fImageAspectRatio
    elif fPaneAspectRatio < fImageAspectRatio:
        fDestHeight = iPaneHeight
        fDestWidth = iPaneHeight / fImageAspectRatio
    return int(fDestWidth), int(fDestHeight)

def _check_file(sFileName):
    """Check if file exists and is readable"""
    bRes = True
    try:
        fTest = file(sFileName, 'rb')
        fTest.close()
    except IOError:
        bRes = False
    return bRes

def _unaccent(sCardName):
    """Remove unicode accents."""
    # Inspired by a post by renaud turned up by google
    # Unicode Normed decomposed form
    sNormed = unicodedata.normalize('NFD',
            unicode(sCardName.encode('utf8'), encoding='utf-8'))
    # Drop non-ascii characters
    return "".join(b for b in sNormed.encode('utf8') if ord(b) < 128)

class CardImagePopupMenu(gtk.Menu):
    # pylint: disable-msg=R0904
    # R0904 - can't not trigger these warning with pygtk
    """Popup menu for the Card Image Frame"""

    def __init__(self, oFrame, iZoomMode):
        super(CardImagePopupMenu, self).__init__()
        self.oFrame = oFrame
        self.oZoom = gtk.RadioAction('Zoom', 'Show images at original size',
                None, None, _iFULL)
        self.oZoom.set_group(None)
        self.oViewFixed = gtk.RadioAction('ViewFixed',
                'Show images at fixed size', None, None, _iVIEW_FIXED)
        self.oViewFixed.set_group(self.oZoom)
        self.oViewFit = gtk.RadioAction('ViewFit', 'Fit images to the pane',
                None, None, _iFIT)
        self.oViewFit.set_group(self.oZoom)
        self.oNext = gtk.Action('NextExp', 'Show next expansion image', None,
                None)
        self.oPrev = gtk.Action('PrevExp', 'Show previous expansion image',
                None, None)

        self.oPrev.connect('activate', self.cycle_expansion, _iBACKWARD)
        self.oNext.connect('activate', self.cycle_expansion, _iFORWARD)
        self.oViewFit.connect('activate', self.set_zoom, _iFIT)
        self.oZoom.connect('activate', self.set_zoom, _iFULL)
        self.oViewFixed.connect('activate', self.set_zoom, _iVIEW_FIXED)

        if iZoomMode == _iFULL:
            self.oZoom.set_active(True)
        elif iZoomMode == _iVIEW_FIXED:
            self.oViewFixed.set_active(True)
        elif iZoomMode == _iFIT:
            self.oViewFit.set_active(True)

        self.add(self.oViewFit.create_menu_item())
        self.add(self.oViewFixed.create_menu_item())
        self.add(self.oZoom.create_menu_item())
        self.add(self.oNext.create_menu_item())
        self.add(self.oPrev.create_menu_item())

    def set_show_expansion_state(self, bValue):
        """Grey out the expansion menus if needed"""
        self.oNext.set_sensitive(bValue)
        self.oPrev.set_sensitive(bValue)

    # pylint: disable-msg=W0613
    # oWidget needed by gtk function signature
    def cycle_expansion(self, oWidget, iDir):
        """Change the expansion as requested."""
        assert(iDir in [_iBACKWARD, _iFORWARD])
        self.oFrame.do_cycle_expansion(iDir)

    def set_zoom(self, oWidget, iScale):
        """Change the drawing mode."""
        assert(iScale in [_iFULL, _iVIEW_FIXED, _iFIT])
        self.oFrame.set_zoom_mode(iScale)

class CardImageFrame(BasicFrame, CardListViewListener):
    # pylint: disable-msg=R0904, R0902
    # R0904 - can't not trigger these warning with pygtk
    # R0902 - we need to keep quite a lot of internal state
    """Frame which displays the image.

       We wrap a gtk.Image in an EventBox (for focus & DnD events)
       and a Viewport (for scrolling)
       """

    def __init__(self, oMainWindow, oConfigFile):
        super(CardImageFrame, self).__init__(oMainWindow)
        self._oConfigFile = oConfigFile
        oVBox = gtk.VBox(False, 2)
        oBox = gtk.EventBox()
        self.oExpansionLabel = gtk.Label()
        oVBox.pack_start(self.oExpansionLabel, False, False)
        oVBox.pack_start(oBox)
        self._oView = AutoScrolledWindow(oVBox, True)
        self._oView.get_hadjustment().connect('changed', self._pane_adjust)
        self._oView.get_vadjustment().connect('changed', self._pane_adjust)
        self._oImage = gtk.Image()
        self._oImage.set_from_stock(gtk.STOCK_MISSING_IMAGE,
                gtk.ICON_SIZE_DIALOG)
        oBox.add(self._oImage)

        # Enable DnD handling, same as for BasicFrame
        aDragTargets = [ ('STRING', 0, 0),
                ('text/plain', 0, 0) ]
        oBox.drag_dest_set(gtk.DEST_DEFAULT_ALL,
                aDragTargets,
                gtk.gdk.ACTION_COPY | gtk.gdk.ACTION_MOVE)
        oBox.connect('drag-data-received', self.drag_drop_handler)
        oBox.connect('drag-motion', self.drag_motion)
        oBox.connect('button-press-event', self.__cycle_expansion)

        self.__sPrefsPath = self._oConfigFile.get_plugin_key('card image path')
        if self.__sPrefsPath is None:
            self.__sPrefsPath = os.path.join(prefs_dir('Sutekh'), 'cardimages')
            self._oConfigFile.set_plugin_key('card image path',
                    self.__sPrefsPath)
        self.__bShowExpansions = self.__have_expansions()
        self.__sCurExpansion = ''
        self.__aExpansions = []
        self.__iExpansionPos = 0
        self.__sCardName = ''
        self.__iZoomMode = _iFIT
        self._tPaneSize = (0, 0)

    type = property(fget=lambda self: "Card Image Frame", doc="Frame Type")

    def __have_expansions(self, sTestPath=''):
        """Test if directory contains expansion/image structure used by ARDB"""
        if sTestPath == '':
            sTestFile = os.path.join(self.__sPrefsPath, 'bh', 'acrobatics.jpg')
        else:
            sTestFile = os.path.join(sTestPath, 'bh', 'acrobatics.jpg')
        return _check_file(sTestFile)

    def __convert_expansion(self, sExpansionName):
        """Convert the Full Expansion name into the abbrevation needed."""
        if sExpansionName == '' or not self.__bShowExpansions:
            return ''
        # pylint: disable-msg=E1101
        # pylint doesn't pick up IExpansion methods correctly
        oExpansion = IExpansion(sExpansionName)
        return oExpansion.shortname.lower()

    def __set_expansion_info(self, sCardName):
        """Set the expansion info."""
        # pylint: disable-msg=E1101
        # pylint doesn't pick up IAbstractCard methods correctly
        try:
            oAbsCard = IAbstractCard(sCardName)
            bHasInfo = len(oAbsCard.rarity) > 0
        except SQLObjectNotFound:
            bHasInfo = False
        if bHasInfo:
            aExp = [oP.expansion.name for oP in oAbsCard.rarity]
            self.__aExpansions = sorted(list(set(aExp))) # remove duplicates
            self.__iExpansionPos = 0
            self.__sCurExpansion = self.__aExpansions[0]
        else:
            self.__sCurExpansion = ''
            self.__aExpansions = []
            self.__iExpansionPos = 0

    def __redraw(self, bPause):
        """Redraw the current card"""
        # If further events are pending, don't try and redraw
        if bPause and gtk.gdk.events_pending():
            return
        sFullFilename = self.__convert_cardname()
        self.__load_image(sFullFilename)

    def __convert_cardname(self):
        """Convert sCardName to the form used by the card image list"""
        sCurExpansionPath = self.__convert_expansion(self.__sCurExpansion)
        sFilename = _unaccent(self.__sCardName.lower())
        if sFilename.find('the ') == 0:
            sFilename = sFilename[4:] + 'the'
        sFilename = sFilename.replace('(advanced)', 'adv')
        # Should probably do this via translate
        for sChar in [" ", ".", ",", "'", "(", ")", "-", ":", "!", '"', "/"]:
            sFilename = sFilename.replace(sChar, '')
        sFilename = sFilename + '.jpg'
        return os.path.join(self.__sPrefsPath, sCurExpansionPath, sFilename)

    def __load_image(self, sFullFilename):
        """Load an image into the pane, show broken image if needed"""
        self._oImage.set_alignment(0.5, 0.5) # Centre image
        try:
            if self.__bShowExpansions:
                self.oExpansionLabel.set_markup('<i>Image from expansion :'
                        ' </i> %s' % self.__sCurExpansion)
                self.oExpansionLabel.show()
                # pylint: disable-msg=E1101
                # pylint doesn't pick up allocation methods correctly
                iHeightOffset = self.oExpansionLabel.allocation.height + 2
            else:
                self.oExpansionLabel.hide() # config chanes can cause this
                iHeightOffset = 0
            oPixbuf = gtk.gdk.pixbuf_new_from_file(sFullFilename)
            iWidth = oPixbuf.get_width()
            iHeight = oPixbuf.get_height()
            if self.__iZoomMode == _iFIT:
                # Need to fix aspect ratios
                iPaneHeight = self._oView.get_vadjustment().page_size - \
                        iHeightOffset
                iPaneWidth = self._oView.get_hadjustment().page_size
                # don't centre image under label
                self._oImage.set_alignment(0, 0.5)
                iDestWidth, iDestHeight = _scale_dims(iWidth, iHeight,
                        iPaneWidth, iPaneHeight)
                if iDestWidth > 0 and iDestHeight > 0:
                    self._oImage.set_from_pixbuf(oPixbuf.scale_simple(
                        iDestWidth, iDestHeight, gtk.gdk.INTERP_HYPER))
                    self._tPaneSize = (self._oView.get_hadjustment().page_size,
                            self._oView.get_vadjustment().page_size)
            elif self.__iZoomMode == _iVIEW_FIXED:
                iDestWidth, iDestHeight = _scale_dims(iWidth, iHeight,
                        _tRATIO[0], _tRATIO[1])
                self._oImage.set_from_pixbuf(oPixbuf.scale_simple(iDestWidth,
                    iDestHeight, gtk.gdk.INTERP_HYPER))
            else:
                # Full size, so no scaling
                self._oImage.set_from_pixbuf(oPixbuf)

        except gobject.GError:
            self._oImage.set_from_stock(gtk.STOCK_MISSING_IMAGE,
                    gtk.ICON_SIZE_DIALOG)
        self._oImage.queue_draw()

    def check_images(self, sTestPath=''):
        """Check if dir contains images in the right structure"""
        self.__bShowExpansions = self.__have_expansions(sTestPath)
        if self.__bShowExpansions:
            return True
        if sTestPath == '':
            sTestFile = os.path.join(self.__sPrefsPath, 'acrobatics.jpg')
        else:
            sTestFile = os.path.join(sTestPath, 'acrobatics.jpg')
        return _check_file(sTestFile)

    def update_config_path(self, sNewPath):
        """Update the path we use to search for expansions."""
        self.__sPrefsPath = sNewPath
        self._oConfigFile.set_plugin_key('card image path', sNewPath)
        self.__bShowExpansions = self.__have_expansions()

    def set_card_text(self, sCardName, sExpansionName=''):
        """Set the image in response to a set card name event."""
        if sCardName != self.__sCardName:
            self.__set_expansion_info(sCardName)
            self.__sCardName = sCardName
        if len(self.__aExpansions) > 0:
            if sExpansionName in self.__aExpansions:
                # Honour expansion from set_card_text
                self.__sCurExpansion = sExpansionName
                self.__iExpansionPos = self.__aExpansions.index(sExpansionName)
            elif self.__sCurExpansion == '' or \
                    self.__sCurExpansion not in self.__aExpansions:
                # Set self.__sCurExpansion to a valid value
                self.__sCurExpansion = self.__aExpansions[0]
                self.__iExpansionPos = 0
        self.__redraw(False)

    def do_cycle_expansion(self, iDir):
        """Change the expansion image to a different one in the list."""
        if len(self.__aExpansions) < 2 or not self.__bShowExpansions:
            return # nothing to scroll through
        if iDir == _iFORWARD:
            self.__iExpansionPos += 1
            if self.__iExpansionPos >= len(self.__aExpansions):
                self.__iExpansionPos = 0
        elif iDir == _iBACKWARD:
            self.__iExpansionPos -= 1
            if self.__iExpansionPos < 0:
                self.__iExpansionPos = len(self.__aExpansions) - 1
        self.__sCurExpansion = self.__aExpansions[self.__iExpansionPos]
        self.__redraw(False)

    def set_zoom_mode(self, iScale):
        """Update the zoom mode."""
        self.__iZoomMode = iScale
        self.__redraw(False)

    # pylint: disable-msg=W0613
    # oWidget needed by gtk function signature
    def __cycle_expansion(self, oWidget, oEvent):
        """On a button click, move to the next expansion."""
        if oEvent.type != gtk.gdk.BUTTON_PRESS:
            return True # don't jump twice on double or triple clicks
        if oEvent.button == 1:
            self.do_cycle_expansion(_iFORWARD)
        elif oEvent.button == 3:
            # Do context menu
            oPopupMenu = CardImagePopupMenu(self, self.__iZoomMode)
            oPopupMenu.set_show_expansion_state(self.__bShowExpansions and
                    len(self.__aExpansions) > 1)
            oPopupMenu.popup(None, None, None, oEvent.button, oEvent.time)
        return True

    # W0613 oAdjust also required by signature
    def _pane_adjust(self, oAdjust):
        """Redraw the image if needed when the pane size changes."""
        if self.__iZoomMode == _iFIT:
            tCurSize = (self._oView.get_hadjustment().page_size,
                self._oView.get_vadjustment().page_size)
            if tCurSize[0] != self._tPaneSize[0] or \
                    tCurSize[1] != self._tPaneSize[1]:
                self.__redraw(True)

# pylint: disable-msg=R0904
# R0904 - gtk Widget, so has many public methods
class ImageConfigDialog(SutekhDialog):
    """Dialog for configuring the Image plugin."""
    def __init__(self, oParent, bFirstTime=False):
        super(ImageConfigDialog, self).__init__('Configure Card Images Plugin',
                oParent, gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                (gtk.STOCK_OK, gtk.RESPONSE_OK, gtk.STOCK_CANCEL,
                    gtk.RESPONSE_CANCEL))
        oDescLabel = gtk.Label()
        if not bFirstTime:
            oDescLabel.set_markup('<b>Choose how to configure the cardimages '
                    'plugin</b>')
        else:
            oDescLabel.set_markup('<b>Choose how to configure the cardimages '
                    'plugin</b>\nChoose cancel to skip configuring the '
                    'images plugin\nYou will not be prompted again')
        self.oChoiceDownload = gtk.RadioButton(
                label='Download and install cardimages from feldb.extra.hu')
        self.oChoiceLocalCopy = gtk.RadioButton(self.oChoiceDownload,
                label='Install images from a local zip file')
        self.oChoiceNewDir = gtk.RadioButton(self.oChoiceDownload,
                label='Choose a directory containing the images')
        oChoiceBox = gtk.VBox(False, 2)
        self.oDirChoiceWidget = gtk.FileChooserWidget(
                action=gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
        self.oFileChoiceWidget = gtk.FileChooserWidget(
                action=gtk.FILE_CHOOSER_ACTION_OPEN)
        self.oChoiceDownload.connect('toggled', self._radio_toggled,
                oChoiceBox, None)
        self.oChoiceLocalCopy.connect('toggled', self._radio_toggled,
                oChoiceBox, self.oFileChoiceWidget)
        self.oChoiceNewDir.connect('toggled', self._radio_toggled, oChoiceBox,
                self.oDirChoiceWidget)

        self.oChoiceDownload.set_active(True)
        # pylint: disable-msg=E1101
        # pylint doesn't pick up vbox methods correctly
        self.vbox.pack_start(oDescLabel, False, False)
        self.vbox.pack_start(self.oChoiceDownload, False, False)
        self.vbox.pack_start(self.oChoiceLocalCopy, False, False)
        self.vbox.pack_start(self.oChoiceNewDir, False, False)
        self.vbox.pack_start(oChoiceBox)
        self.set_size_request(600, 600)

        #oDialog.vbox.pack_start(oFileWidget)
        sPrefsPath = oParent.config_file.get_plugin_key('card image path')
        if sPrefsPath is not None:
            if os.path.exists(sPrefsPath) and os.path.isdir(sPrefsPath):
                # File widget doesn't like being pointed at non-dirs
                # when in SELECT_FOLDER mode
                self.oDirChoiceWidget.set_current_folder(sPrefsPath)
        oZipFilter = gtk.FileFilter()
        oZipFilter.add_pattern('*.zip')
        oZipFilter.add_pattern('*.ZIP')
        self.oFileChoiceWidget.add_filter(oZipFilter)
        self.show_all()

    def _radio_toggled(self, oRadioButton, oChoiceBox, oFileWidget):
        """Display the correct file widget when radio buttons change"""
        if oRadioButton.get_active():
            for oChild in oChoiceBox.get_children():
                oChoiceBox.remove(oChild)
            if oFileWidget:
                oChoiceBox.pack_start(oFileWidget)
            self.show_all()

    def get_path(self):
        """Get the path from oDirChoiceWidget."""
        return self.oDirChoiceWidget.get_filename()

    def get_file(self):
        """Get the file from oFileChoiceWidget."""
        return self.oFileChoiceWidget.get_filename()

    def get_choice(self):
        """Get which of the RadioButtons was active."""
        if self.oChoiceDownload.get_active():
            return 'Download'
        elif self.oChoiceLocalCopy.get_active():
            return 'Local copy'
        else:
            return 'New directory'

class CardImagePlugin(CardListPlugin):
    """Plugin providing access to CardImageFrame."""
    dTableVersions = {PhysicalCardSet : [5]}
    aModelsSupported = ["MainWindow"]
    aListenViews = [PhysicalCardSet,
            PhysicalCard]

    oImageFrame = None

    _sMenuFlag = 'Card Image Frame'

    # pylint: disable-msg=W0142
    # ** magic OK here
    def __init__(self, *aArgs, **kwargs):
        super(CardImagePlugin, self).__init__(*aArgs, **kwargs)
        # Add listeners to the appropriate views
        if not self.image_frame:
            self.init_image_frame(self)
        if self._cModelType in self.aListenViews:
            self.view.add_listener(self.image_frame)
        self._oReplaceItem = None
        self._oAddItem = None
        self._oConfigMenuItem = None

    image_frame = property(fget=lambda self: self.get_image_frame(),
            doc="The image frame")

    @classmethod
    def get_image_frame(cls):
        """Return the global ImageFrame."""
        return cls.oImageFrame

    @classmethod
    def init_image_frame(cls, oCur):
        """Setup the global image frame."""
        if not cls.oImageFrame:
            cls.oImageFrame = CardImageFrame(oCur.parent,
                    oCur.parent.config_file)
            cls.oImageFrame.set_title(cls._sMenuFlag)
            cls.oImageFrame.add_parts()

    def get_menu_item(self):
        """Overrides method from base class.

           Adds the menu item on the MainWindow if the images can be found.
           """
        if not self.check_versions() or not self.check_model_type():
            return None
        self._oReplaceItem = gtk.MenuItem("Replace with Card Image Frame")
        self._oReplaceItem.connect("activate", self.replace_pane)

        self._oAddItem = gtk.MenuItem("Add Card Image Frame")
        self._oAddItem.connect("activate", self.add_pane)
        self.parent.add_to_menu_list('Card Image Frame',
                self.add_image_frame_active)
        self._oConfigMenuItem = gtk.MenuItem("Configure Card Images Plugin")
        self._oConfigMenuItem.connect("activate", self.config_activate)
        if not self.image_frame.check_images():
            # Don't allow the menu option if we can't find the images
            self.add_image_frame_active(False)
        return [('Plugins', self._oConfigMenuItem),
                ('Add Pane', self._oAddItem),
                ('Replace Pane', self._oReplaceItem)]

    def setup(self):
        """Prompt the user to download/setup images the first time"""
        sPrefsPath = self.parent.config_file.get_plugin_key('card image path')
        if not os.path.exists(sPrefsPath):
            # Looks like the first time
            oDialog = ImageConfigDialog(self.parent, True)
            self.handle_response(oDialog)
            # Path may have been changed, so we need to requery config file
            sPrefsPath = self.parent.config_file.get_plugin_key(
                    'card image path')
            # Don't get called next time
            ensure_dir_exists(sPrefsPath)

    # pylint: disable-msg=W0613
    # oMenuWidget needed by gtk function signature
    def config_activate(self, oMenuWidget):
        """Configure the plugin dialog."""
        oDialog = ImageConfigDialog(self.parent, False)
        self.handle_response(oDialog)

    def handle_response(self, oDialog):
        """Handle the response from the config dialog"""
        iResponse = oDialog.run()
        bActivateMenu = False
        if iResponse == gtk.RESPONSE_OK:
            sChoice = oDialog.get_choice()
            if sChoice == 'Download':
                if self._download_file():
                    bActivateMenu = True
                else:
                    do_complaint_error('Unable to successfully download'
                            ' and install zipfile')
            elif sChoice == 'Local copy':
                sFileName = oDialog.get_file()
                if self._unzip_file(sFileName):
                    bActivateMenu = True
                else:
                    do_complaint_error('Unable to successfully install '
                            ' zipfile %s' % sFileName)
            else:
                # New directory
                sTestPath = oDialog.get_path()
                if self._accept_path(sTestPath):
                    # Update preferences
                    self.image_frame.update_config_path(sTestPath)
                    bActivateMenu = True
        if bActivateMenu:
            # Update the menu display if needed
            if self._sMenuFlag not in self.parent.dOpenFrames.values():
                # Pane is not open, so try to enable menu
                self.add_image_frame_active(True)
        # get rid of the dialog
        oDialog.destroy()

    def _download_file(self):
        """Download a zip file containing the images."""
        # Download and install http://feldb.extra.hu/pictures.zip
        sUrl = 'http://feldb.extra.hu/pictures.zip'
        oFile = urllib2.urlopen(sUrl)
        sDownloadDir = prefs_dir('Sutekh')
        ensure_dir_exists(sDownloadDir)
        fTemp = tempfile.TemporaryFile()
        sLength = oFile.info().getheader('Content-Length')
        if sLength:
            iLength = int(sLength)
            oProgress = ProgressDialog()
            oProgress.set_description('Download progress')
            iTotal = 0
            bCont = True
            while bCont:
                oInf = oFile.read(10000)
                iTotal += 10000
                oProgress.update_bar(float(iTotal) / iLength)
                if oInf:
                    fTemp.write(oInf)
                else:
                    bCont = False
                    oProgress.destroy()
        else:
            # Just try and download
            fTemp.write(oFile.read())
        try:
            oZipFile = zipfile.ZipFile(fTemp, 'r')
            bRet = self._unzip_heart(oZipFile)
        except zipfile.BadZipfile:
            bRet = False
        # cleanup file
        fTemp.close()
        return bRet

    def _unzip_file(self, sFileName):
        """Unzip a file containing the images."""
        if _check_file(sFileName):
            try:
                oZipFile = zipfile.ZipFile(sFileName, 'r')
            except zipfile.BadZipfile:
                return False
            return self._unzip_heart(oZipFile)
        return False

    def _unzip_heart(self, oZipFile):
        """Heavy lifting of unzipping a file"""
        sPrefsPath = self.parent.config_file.get_plugin_key('card image path')
        ensure_dir_exists(sPrefsPath)
        iNumber = len(oZipFile.infolist())
        if iNumber < 300:
            # zipfile too short, so don't bother
            return False
        oProgressDialog = ProgressDialog()
        oProgressDialog.set_description('Unzipping')
        iCur = 0
        for oItem in oZipFile.infolist():
            iCur += 1
            oProgressDialog.update_bar(float(iCur) / iNumber)
            oData = oZipFile.read(oItem.filename)
            # Should I also test for cardimages\ ?
            sFileName = os.path.join(sPrefsPath,
                    oItem.filename.replace('cardimages/',''))
            sFileName = sFileName.replace('/', os.path.sep)
            sDir = os.path.dirname(sFileName)
            ensure_dir_exists(sDir)
            oOutputFile = file(sFileName, 'wb')
            oOutputFile.write(oData)
            oOutputFile.close()
        oProgressDialog.destroy()
        if self.image_frame.check_images(sPrefsPath):
            return True
        return False

    def _accept_path(self, sTestPath):
        """Check if the path from user is OK."""
        if sTestPath is not None:
            # Test if path has images
            if not self.image_frame.check_images(sTestPath):
                iQuery = do_complaint_buttons(
                        "Folder does not seem to contain images\n"
                        "Are you sure?", gtk.MESSAGE_QUESTION,
                        (gtk.STOCK_YES, gtk.RESPONSE_YES,
                            gtk.STOCK_NO, gtk.RESPONSE_NO))
                if iQuery == gtk.RESPONSE_NO:
                    # Treat as cancelling
                    return False
            return True
        return False # No path, can't be OK

    def add_image_frame_active(self, bValue):
        """Toggle the sensitivity of the menu item."""
        if bValue and not self.image_frame.check_images():
            # Can only be set true if check_images returns true
            self._oReplaceItem.set_sensitive(False)
            self._oAddItem.set_sensitive(False)
        else:
            self._oReplaceItem.set_sensitive(bValue)
            self._oAddItem.set_sensitive(bValue)

    def get_frame_from_config(self, sType):
        """Add the frame if it's been saved in the config file."""
        if sType == self._sMenuFlag:
            return (self.image_frame, self._sMenuFlag)
        else:
            return None

    # pylint: disable-msg=W0613
    # oWidget needed by gtk function signature
    def replace_pane(self, oWidget):
        """Handle replacing a frame to the main window if required"""
        if self._sMenuFlag not in self.parent.dOpenFrames.values():
            oNewPane = self.parent.focussed_pane
            if oNewPane:
                self.parent.replace_frame(oNewPane, self.image_frame,
                        self._sMenuFlag)

    def add_pane(self, oWidget):
        """Handle adding the frame to the main window if required"""
        if self._sMenuFlag not in self.parent.dOpenFrames.values():
            oNewPane = self.parent.add_pane_end()
            self.parent.replace_frame(oNewPane, self.image_frame,
                    self._sMenuFlag)


# pylint: disable-msg=C0103
# shut up complaint about the name
plugin = CardImagePlugin



# test_WriteLackeyCCG.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Test Writing a card set to an Lackey CCG file"""

from sutekh.tests.TestCore import SutekhTest
from sutekh.tests.core.test_PhysicalCardSet import aCardSetNames, \
        get_phys_cards
from sutekh.core.SutekhObjects import PhysicalCardSet
from sutekh.io.WriteLackeyCCG import  WriteLackeyCCG
import unittest

sExpected1 = """4\t.44 Magnum
2\tAK-47
2\tAbbot
2\tAbombwe
1\tPath of Blood, The
Crypt:
1\tAbebe
1\tAlan Sovereign Adv.
"""

class LackeyWriterTests(SutekhTest):
    """class for the Lackey CCG deck writer tests"""

    def test_deck_writer(self):
        """Test Lackey CCG deck writing"""
        # pylint: disable-msg=E1101, R0915, R0914
        # E1101: SQLObject + PyProtocols magic confuses pylint
        # R0915, R0914: Want a long, sequential test case to minimise
        # repeated setups, so it has lots of lines + variables
        aAddedPhysCards = get_phys_cards()
        # We have a physical card list, so create some physical card sets
        oPhysCardSet1 = PhysicalCardSet(name=aCardSetNames[0])
        oPhysCardSet1.comment = 'A test comment'
        oPhysCardSet1.author = 'A test author'

        for oCard in aAddedPhysCards:
            oPhysCardSet1.addPhysicalCard(oCard.id)
            oPhysCardSet1.syncUpdate()
        oPhysCardSet1.addPhysicalCard(aAddedPhysCards[0])
        oPhysCardSet1.addPhysicalCard(aAddedPhysCards[0])
        oPhysCardSet1.syncUpdate()

        # Check output

        oWriter = WriteLackeyCCG()
        sTempFileName =  self._create_tmp_file()
        fOut = file(sTempFileName, 'w')
        oWriter.write(fOut, oPhysCardSet1)
        fOut.close()

        fIn = open(sTempFileName, 'rU')
        sData = fIn.read()
        fIn.close()

        self.assertEqual(sData, sExpected1)


if __name__ == "__main__":
    unittest.main()



# CardSetController.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2006 Neil Muller <drnlmuller+sutekh@gmail.com>
# Copyright 2006 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details

"""Controller for the card sets"""

from sqlobject import SQLObjectNotFound
from sutekh.gui.CardSetManagementController import check_ok_to_delete, \
        update_card_set
from sutekh.gui.CardSetView import CardSetView
from sutekh.gui.CreateCardSetDialog import CreateCardSetDialog
from sutekh.core.DBSignals import listen_row_destroy, listen_row_update, \
        send_changed_signal, disconnect_row_destroy, disconnect_row_update
from sutekh.core.SutekhObjects import IPhysicalCardSet, PhysicalCardSet, \
        AbstractCard, PhysicalCard, MapPhysicalCardToPhysicalCardSet, \
        IExpansion, IPhysicalCard
from sutekh.core.CardSetUtilities import delete_physical_card_set

class CardSetController(object):
    """Controller class for the Card Sets."""
    _sFilterType = 'PhysicalCard'

    def __init__(self, sName, oMainWindow, oFrame):
        # pylint: disable-msg=E1101
        # SQLObject methods confuse pylint
        self._oMainWindow = oMainWindow
        self._oMenu = None
        self._oFrame = oFrame
        self._oView = CardSetView(oMainWindow, self, sName)
        self.__oPhysCardSet = IPhysicalCardSet(sName)
        # listen on card set signals
        # We listen here, rather than in the model (as for card set changes),
        # to use queue_reload to delay processing until after the database
        listen_row_update(self.card_set_changed, PhysicalCardSet)
        listen_row_destroy(self.card_set_deleted, PhysicalCardSet)
        # We don't listen for card set creation, since newly created card
        # sets aren't inuse. If that changes, we'll need to add an additional
        # signal listen here

    # pylint: disable-msg=W0212
    # explicitly allow access to these values via thesep properties
    view = property(fget=lambda self: self._oView, doc="Associated View")
    model = property(fget=lambda self: self._oView._oModel, doc="View's Model")
    frame = property(fget=lambda self: self._oFrame, doc="Associated Frame")
    filtertype = property(fget=lambda self: self._sFilterType,
            doc="Associated Type")
    # pylint: enable-msg=W0212

    def cleanup(self):
        """Remove the signal handlers."""
        disconnect_row_update(self.card_set_changed, PhysicalCardSet)
        disconnect_row_destroy(self.card_set_deleted, PhysicalCardSet)
        self.model.cleanup()

    def card_set_changed(self, oCardSet, dChanges):
        """When changes happen that may effect this card set, reload.

           Cases are:
             * when the parent changes
             * when a child card set is added or removed while marked in use
             * when a child card set is marked/unmarked as in use.
             * when a 'sibling' card set is marked/unmarked as in use
             * when a 'sibling' card set is added or removed while marked in
               use
           Whether this requires a reload depends on the current model mode.
           When the parent changes to or from none, we also update the menus
           and the parent card shown view.
           """
        # pylint: disable-msg=E1101
        # Pyprotocols confuses pylint
        if oCardSet.id == self.__oPhysCardSet.id and \
                dChanges.has_key('parentID'):
            # This card set's parent is changing
            # Update menu to reflect this
            self._oFrame.menu.check_parent_count_column(oCardSet.parent,
                    dChanges['parentID'])
            if self.model.changes_with_parent() or not dChanges['parentID']:
                # Parent count is shown, or not shown becuase parent is
                # changing to None, so this affects the shown cards.
                self._oFrame.queue_reload()
        elif oCardSet.parent and oCardSet.parent.id == self.__oPhysCardSet.id \
                and self.model.changes_with_children():
            # This is a child card set, and this can require a reload
            if dChanges.has_key('inuse'):
                # inuse flag being toggled
                self._oFrame.queue_reload()
            elif dChanges.has_key('parentID') and oCardSet.inuse:
                # Inuse card set is being reparented
                self._oFrame.queue_reload()
        elif dChanges.has_key('parentID') and \
                dChanges['parentID'] == self.__oPhysCardSet.id and \
                oCardSet.inuse and self.model.changes_with_children():
            # acquiring a new inuse child card set
            self._oFrame.queue_reload()
        elif self.__oPhysCardSet.parent and self.model.changes_with_siblings():
            # Sibling's are possible, so check for them
            if dChanges.has_key('ParentID') and oCardSet.inuse:
                # Possibling acquiring or losing inuse sibling
                if (dChanges['ParentID'] == self.__oPhysCardSet.parent.id) or \
                        (oCardSet.parent and oCardSet.parent.id ==
                                self.__oPhysCardSet.parent.id):
                    # Reload if needed
                    self._oFrame.queue_reload()
            elif dChanges.has_key('inuse') and oCardSet.parent and \
                    oCardSet.parent.id == self.__oPhysCardSet.parent.id:
                # changing inuse status of sibling
                self._oFrame.queue_reload()

    # pylint: disable-msg=W0613
    # fPostFuncs is passed by SQLObject 0.10, but not by 0.9
    def card_set_deleted(self, oCardSet, fPostFuncs=None):
        """Listen for card set removal events.

           Needed if child card sets are deleted, for instance.
           """
        # pylint: disable-msg=E1101
        # Pyprotocols confuses pylint
        if oCardSet.parent and oCardSet.parent.id == \
                self.__oPhysCardSet.id and oCardSet.inuse and \
                self.model.changes_with_children():
            # inuse child card set going, so we need to reload
            self._oFrame.queue_reload()
        if self.__oPhysCardSet.parent and self.model.changes_with_siblings() \
                and oCardSet.parent and oCardSet.inuse and \
                oCardSet.parent.id == self.__oPhysCardSet.parent.id:
            # inuse sibling card set going away while this affects display,
            # so reload
            self._oFrame.queue_reload()
        # Other card set deletions don't need to be watched here, since the
        # fiddling on parents should generate changed signals for us.

    def toggle_icons(self, oWidget):
        """Toggle the icons display"""
        self.model.bUseIcons = oWidget.active
        self.view.reload_keep_expanded()

    def set_card_text(self, sCardName):
        """Set card text to reflect selected card."""
        self._oMainWindow.set_card_text(sCardName)

    def inc_card(self, sName, sExpansion, sCardSetName):
        """Returns True if a card was successfully added, False otherwise."""
        return self.add_card(sName, sExpansion, sCardSetName)

    def dec_card(self, sName, sExpansion, sCardSetName):
        """Returns True if a card was successfully removed, False otherwise."""
        # pylint: disable-msg=E1101
        # SQLObject methods confuse pylint
        try:
            if sCardSetName:
                oThePCS = IPhysicalCardSet(sCardSetName)
            else:
                oThePCS = self.__oPhysCardSet
            oAbsCard = AbstractCard.byCanonicalName(sName.lower())

            # find if there's a physical card of that name in the Set
            if not sExpansion:
                # Not expansion specified, so consider all physical cards
                aPhysCards = list(PhysicalCard.selectBy(
                    abstractCardID=oAbsCard.id))
            else:
                if sExpansion == self.model.sUnknownExpansion:
                    oExp = None
                else:
                    oExp = IExpansion(sExpansion)
                aPhysCards = [IPhysicalCard((oAbsCard, oExp))]
        except SQLObjectNotFound:
            # Bail on error
            return False

        for oCard in aPhysCards:
            # Need to remove a single physical card from the mapping table
            # Can't use PhysicalCardSet.remove, as that removes all the cards
            aCandCards = list(MapPhysicalCardToPhysicalCardSet.selectBy(
                    physicalCardID=oCard.id,
                    physicalCardSetID=oThePCS.id))
            if len(aCandCards) > 0:
                # Found candidates, so remove last one
                MapPhysicalCardToPhysicalCardSet.delete(aCandCards[-1].id)
                oThePCS.syncUpdate()
                # signal to update the model
                send_changed_signal(oThePCS, oCard, -1)
                return True
        # Got here, so we failed to remove a card
        return False

    def add_card(self, sName, sExpansion, sCardSetName):
        """Returns True if a card was successfully added, False otherwise."""
        # pylint: disable-msg=E1101
        # SQLObject methods confuse pylint
        try:
            if sCardSetName:
                oThePCS = IPhysicalCardSet(sCardSetName)
            else:
                oThePCS = self.__oPhysCardSet
            oAbsCard = AbstractCard.byCanonicalName(sName.lower())

            if not sExpansion or sExpansion == self.model.sUnknownExpansion:
                oExp = None
            else:
                oExp = IExpansion(sExpansion)
            oCard = IPhysicalCard((oAbsCard, oExp))
        except SQLObjectNotFound:
            # Any error means we bail
            return False

        oThePCS.addPhysicalCard(oCard.id)
        oThePCS.syncUpdate()
        # Signal to update the model
        send_changed_signal(oThePCS, oCard, 1)
        return True

    def edit_properties(self, oMenu):
        """Run the dialog to update the card set properties"""
        # pylint: disable-msg=E1101
        # pyprotocols confuses pylint
        oProp = CreateCardSetDialog(self._oMainWindow,
                oCardSet=self.__oPhysCardSet)
        oProp.run()
        sName = oProp.get_name()
        if sName:
            if sName != self.view.sSetName:
                self.view.update_name(sName)
            update_card_set(self.__oPhysCardSet, oProp, self._oMainWindow,
                    self._oFrame.menu)

    def update_to_new_db(self):
        """Update the internal card set to the new DB."""
        try:
            self.__oPhysCardSet = IPhysicalCardSet(self.view.sSetName)
            self.model.update_to_new_db(self.view.sSetName)
        except SQLObjectNotFound:
            # No longer in the database, so remove from the window
            self._oFrame.close_frame()

    def delete_card_set(self):
        """Delete this card set from the database."""
        # Check if CardSet is empty
        # pylint: disable-msg=E1101
        # sqlobject confuses pylint
        if check_ok_to_delete(self.__oPhysCardSet):
            delete_physical_card_set(self.view.sSetName)
            # Tell window to clean up
            # Card Set was deleted, so close up
            self._oFrame.close_frame()

    def add_paste_data(self, sSource, aCards):
        """Helper function for drag+drop and copy+paste.

           Only works when we're editable.
           """
        aSources = sSource.split(':')
        if not self.model.bEditable:
            return False
        if aSources[0] in ["Phys", PhysicalCardSet.sqlmeta.table]:
            # Add the cards, Count Matters
            for iCount, sCardName, sExpansion in aCards:
                # pylint: disable-msg=W0612
                # iLoop is just loop counter
                # Use None to indicate this card set
                if aSources[0] == "Phys":
                    # Only ever add 1 when dragging from physiscal card list
                    self.add_card(sCardName, sExpansion, None)
                else:
                    for iLoop in range(iCount):
                        self.add_card(sCardName, sExpansion, None)
            return True
        else:
            return False

    def del_selected_cards(self, dSelectedData):
        """Helper function to delete the selected data."""
        for sCardName in dSelectedData:
            for sExpansion, iCount in dSelectedData[sCardName].iteritems():
                # pylint: disable-msg=W0612
                # iAttempt is loop counter
                for iAttempt in range(iCount):
                    # None as card set indicates this card set
                    if sExpansion != 'None':
                        self.dec_card(sCardName, sExpansion, None)
                    else:
                        self.dec_card(sCardName, None, None)

    def set_selected_card_count(self, dSelectedData, iNewCnt):
        """Helper function to set the selected cards to the specified number"""
        for sCardName in dSelectedData:
            for (sExpansion, sCardSetName), iCardCount in \
                    dSelectedData[sCardName].iteritems():
                # pylint: disable-msg=W0612
                # iAttempt is loop counter
                if iNewCnt < iCardCount:
                    # remove cards
                    for iAttempt in range(iCardCount - iNewCnt):
                        # None as card set indicates this card set
                        if sExpansion != 'None' and sExpansion != 'All':
                            self.dec_card(sCardName, sExpansion, sCardSetName)
                        else:
                            self.dec_card(sCardName, None, sCardSetName)
                elif iNewCnt > iCardCount:
                    # add cards
                    for iAttempt in range(iNewCnt - iCardCount):
                        # None as card set indicates this card set
                        if sExpansion != 'None' and sExpansion != 'All':
                            self.inc_card(sCardName, sExpansion, sCardSetName)
                        else:
                            self.inc_card(sCardName, None, sCardSetName)

    def alter_selected_card_count(self, dSelectedData, iChg):
        """Helper function to inc/dec the count for the selected cards"""
        for sCardName in dSelectedData:
            for (sExpansion, sCardSetName) in dSelectedData[sCardName]:
                # None as card set in inc_card/dec_Card indicates this card set
                if iChg == -1:
                    # remove cards
                    if sExpansion != 'None' and sExpansion != 'All':
                        self.dec_card(sCardName, sExpansion, sCardSetName)
                    else:
                        self.dec_card(sCardName, None, sCardSetName)
                elif iChg == 1:
                    # add cards
                    if sExpansion != 'None' and sExpansion != 'All':
                        self.inc_card(sCardName, sExpansion, sCardSetName)
                    else:
                        self.inc_card(sCardName, None, sCardSetName)



# CardSetManagementController.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Controller class the card set list."""

import gtk
from sqlobject import SQLObjectNotFound
from sutekh.core.SutekhObjects import PhysicalCardSet, IPhysicalCardSet
from sutekh.gui.FilterDialog import FilterDialog
from sutekh.gui.CardSetManagementView import CardSetManagementView
from sutekh.gui.CreateCardSetDialog import CreateCardSetDialog
from sutekh.gui.SutekhDialog import do_complaint_warning, do_complaint
from sutekh.core.CardSetUtilities import delete_physical_card_set, \
        find_children, detect_loop, get_loop_names, break_loop

def split_selection_data(sSelectionData):
    """Helper function to subdivide selection string into bits again"""
    if sSelectionData == '':
        return 'None', ['']
    aLines = sSelectionData.splitlines()
    sSource = aLines[0]
    if sSource == "Sutekh Pane:" or sSource == 'Card Set:':
        return sSource, aLines
    # Irrelevant to us
    return 'None', ['']

def reparent_card_set(oCardSet, oNewParent):
    """Helper function to ensure that reparenting a card set doesn't
       cause loops"""
    if oNewParent:
        # Can only be a problem if the new parent is a card set
        oOldParent = oCardSet.parent
        oCardSet.parent = oNewParent
        oCardSet.syncUpdate()
        if detect_loop(oCardSet):
            oCardSet.parent = oOldParent
            oCardSet.syncUpdate()
            do_complaint('Changing parent of %s to %s introduces a'
                    ' loop. Leaving the parent unchanged.' % (oCardSet.name,
                        oNewParent.name), gtk.MESSAGE_WARNING,
                    gtk.BUTTONS_CLOSE)
        else:
            return True
    else:
        oCardSet.parent = oNewParent
        oCardSet.syncUpdate()
        return True
    return False

def check_ok_to_delete(oCardSet):
    """Check if the user is OK with deleting the card set."""
    aChildren = find_children(oCardSet)
    iResponse = gtk.RESPONSE_OK
    if len(oCardSet.cards) > 0 and len(aChildren) > 0:
        iResponse = do_complaint_warning("Card Set %s Not Empty and"
                " Has Children. Really Delete?" % oCardSet.name)
    elif len(oCardSet.cards) > 0:
        iResponse = do_complaint_warning("Card Set %s Not Empty."
                " Really Delete?" % oCardSet.name)
    elif len(aChildren) > 0:
        iResponse = do_complaint_warning("Card Set %s"
                " Has Children. Really Delete?" % oCardSet.name)
    return iResponse == gtk.RESPONSE_OK

def update_card_set(oCardSet, oEditDialog, oMainWindow, oMenu):
    """Update the details of the card set when the user edits them."""
    sOldName = oCardSet.name
    sName = oEditDialog.get_name()
    if sName != sOldName:
        oCardSet.name = sName
        oMainWindow.rename_pane(sOldName, sName)
    oCardSet.author = oEditDialog.get_author()
    oCardSet.comment = oEditDialog.get_comment()
    oCardSet.annotations = oEditDialog.get_annotations()
    oParent = oEditDialog.get_parent()
    if oParent != oCardSet.parent:
        reparent_card_set(oCardSet, oParent)
    oCardSet.syncUpdate()
    # Update frame menu
    if oMenu:
        oMenu.update_card_set_menu(oCardSet)
    oMainWindow.reload_pcs_list()

class CardSetManagementController(object):
    """Controller object for the card set list."""
    # pylint: disable-msg=R0904
    # gtk.Widget, so lots of public methods
    def __init__(self, oMainWindow, oFrame):
        self._oMainWindow = oMainWindow
        self._oFrame = oFrame
        self._oFilterDialog = None
        # Check the current set for loops
        for oCS in PhysicalCardSet.select():
            if detect_loop(oCS):
                sLoop = "->".join(get_loop_names(oCS))
                sBreakName = break_loop(oCS)
                do_complaint(
                        'Loop %s in the card sets relationships.\n'
                        'Breaking at %s' % (sLoop, sBreakName),
                        gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE)
                # We break the loop, and let the user fix things,
                # rather than try and be too clever
        self._oView = CardSetManagementView(oMainWindow, self)
        self._oModel = self._oView.get_model()

    # pylint: disable-msg=W0212
    # explicitly allow access to these values via thesep properties
    view = property(fget=lambda self: self._oView, doc="Associated View")
    model = property(fget=lambda self: self._oModel, doc="View's Model")
    frame = property(fget=lambda self: self._oFrame, doc="Associated Frame")
    filtertype = property(fget=lambda self: self._sFilterType,
            doc="Associated Type")
    # pylint: enable-msg=W0212

    # pylint: disable-msg=W0613
    # oWidget, oMenuItem required by function signature
    def get_filter(self, oMenuItem):
        """Get the Filter from the FilterDialog."""
        if self._oFilterDialog is None:
            self._oFilterDialog = FilterDialog(self._oMainWindow,
                    self._oMainWindow.config_file, 'PhysicalCardSet')

        self._oFilterDialog.run()

        if self._oFilterDialog.was_cancelled():
            return # Change nothing

        oFilter = self._oFilterDialog.get_filter()
        if oFilter != None:
            self._oModel.selectfilter = oFilter
            if not self._oModel.applyfilter:
                # If a filter is set, automatically apply
                oMenuItem.set_apply_filter(True)
            else:
                # Filter Changed, so reload
                self._oModel.load()
        else:
            # Filter is set to blank, so we treat this as disabling
            # Filter
            if self._oModel.applyfilter:
                oMenuItem.set_apply_filter(False)
            else:
                self._oModel.load()

    def run_filter(self, bState):
        """Enable or diable the current filter based on bState"""
        if self._oModel.applyfilter != bState:
            self._oModel.applyfilter = bState
            self._oModel.load()

    def create_new_card_set(self, oWidget):
        """Create a new card set"""
        oDialog = CreateCardSetDialog(self._oMainWindow)
        oDialog.run()
        sName = oDialog.get_name()
        # pylint: disable-msg=E1102, W0612
        # W0612 - oCS isn't important, as the creation of the new card
        # set is what matters
        if sName:
            sAuthor = oDialog.get_author()
            sComment = oDialog.get_comment()
            oParent = oDialog.get_parent()
            oCS = PhysicalCardSet(name=sName, author=sAuthor,
                    comment=sComment, parent=oParent)
            self._oMainWindow.add_new_physical_card_set(sName)

    def edit_card_set_properties(self, oWidget):
        """Create a new card set"""
        sSetName = self._oView.get_selected_card_set()
        if not sSetName:
            return
        oFrame = self._oMainWindow.find_pane_by_name(sSetName)
        if oFrame:
            oMenu = oFrame.menu
        else:
            oMenu = None
        oCardSet = IPhysicalCardSet(sSetName)
        oDialog = CreateCardSetDialog(self._oMainWindow, oCardSet=oCardSet)
        oDialog.run()
        sName = oDialog.get_name()
        if sName:
            if oFrame and sName != oFrame.view.sSetName:
                oFrame.view.update_name(sName)
            update_card_set(oCardSet, oDialog, self._oMainWindow,
                    oMenu)

    def delete_card_set(self, oWidget):
        """Delete the selected card set."""
        sSetName = self._oView.get_selected_card_set()
        if not sSetName:
            return
        # pylint: disable-msg=E1101
        # sqlobject confuses pylint
        try:
            oCS = PhysicalCardSet.byName(sSetName)
        except SQLObjectNotFound:
            return
        # Check for card sets this is a parent of
        if check_ok_to_delete(oCS):
            sFrameName = sSetName
            delete_physical_card_set(sSetName)
            self._oMainWindow.remove_frame_by_name(sFrameName)
            self.reload_keep_expanded(False)

    def toggle_in_use_flag(self, oMenuItem):
        """Toggle the in-use status of the card set"""
        sSetName = self._oView.get_selected_card_set()
        if not sSetName:
            return
        try:
            # pylint: disable-msg=E1101
            # SQLObject confuses pylint
            oCS = PhysicalCardSet.byName(sSetName)
        except SQLObjectNotFound:
            return
        oCS.inuse = not oCS.inuse
        oCS.syncUpdate()
        self.reload_keep_expanded(True)

    def row_clicked(self, oTreeView, oPath, oColumn):
        """Handle row clicked events.

           allow double clicks to open a card set.
           """
        sName = self._oModel.get_name_from_path(oPath)
        # check if card set is open before opening again
        oPane = self._oMainWindow.find_pane_by_name(sName)
        if oPane is not None:
            return # Already open, so do nothing
        self._oMainWindow.add_new_physical_card_set(sName)

    def reload_keep_expanded(self, bRestoreSelection):
        """Reload with current expanded state.

           Attempt to reload the card list, keeping the existing structure
           of expanded rows.
           """
        # Internal helper functions
        # See what's expanded
        # pylint: disable-msg=W0612
        # we're not interested in oModel here
        oSelection = self._oView.get_selection()
        if oSelection is not None:
            oModel, aSelectedRows = oSelection.get_selected_rows()
        else:
            aSelectedRows = []
        if len(aSelectedRows) > 0:
            oSelPath = aSelectedRows[0]
        else:
            oSelPath = None
        dExpandedDict = {}
        self._oModel.foreach(self._oView.get_row_status, dExpandedDict)
        # Reload, but use cached info
        self._oModel.load()
        # Re-expand stuff
        self._oView.set_row_status(dExpandedDict)
        if oSelPath and bRestoreSelection:
            # Restore selection
            oSelection.select_path(oSelPath)
        # Handle any defferred database single issues
        self._oMainWindow.do_all_queued_reloads()

    # pylint: disable-msg=R0913
    # arguments as required by function signature
    def card_set_drop(self, oWdgt, oContext, iXPos, iYPos, oData, oInfo,
            oTime):
        """Default drag-n-drop handler."""
        # Pass off to the Frame Handler
        sSource, aData = split_selection_data(oData.data)
        if sSource == "Sutekh Pane:":
            self._oFrame.drag_drop_handler(oWdgt, oContext, iXPos, iYPos,
                    oData, oInfo, oTime)
        elif sSource == "Card Set:":
            # Find the card set at iXPos, iYPos
            # Need to do this to skip avoid headers and such confusing us
            oPath = self._oView.get_path_at_pointer()
            if oPath:
                sTargetName = self._oModel.get_name_from_path(oPath)
                sThisName = aData[1]
                # pylint: disable-msg=W0704
                # doing nothing on SQLObjectNotFound seems the best choice
                try:
                    oDraggedCS = IPhysicalCardSet(sThisName)
                    oParentCS = IPhysicalCardSet(sTargetName)
                    if reparent_card_set(oDraggedCS, oParentCS):
                        self.reload_keep_expanded(False)
                        oPath = self._oModel.get_path_from_name(sThisName)
                        # Make newly dragged set visible
                        if oPath:
                            self._oView.expand_to_path(oPath)
                        oContext.finish(True, False, oTime)
                except SQLObjectNotFound:
                    pass
        oContext.finish(False, False, oTime)

    # pylint: enable-msg=R0913
